// ##############################################################################
// Baum-welch optimization of transition frequencies
// Copyright Alex Nguyen Ba 2011
// Version 1.2
//
// LICENCE 
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ##############################################################################


using namespace std;

#include <string>
#include <iostream>
#include <fstream>
#include <map>
#include <stdlib.h> 
#include <cstring>
#include <math.h>
#include <vector>

class doubleindex{
public:
	int incsize;
	double* data;
	doubleindex(int sizea, int sizeb){
		incsize = sizea;
		data = new double[sizea*sizeb];
		memset(data,'\0',sizeof(double)*sizea*sizeb);
	}
	~doubleindex(){
	}
	
	double read(int a, int b){return(data[a + b * incsize]);}
	void write(int a, int b, double in ){data[a + b * incsize] =in ;}
	double& operator()(int a, int b){
		return(data[a + b * incsize]);
	}
	
	void clear(){delete[](data);}

};

class singleindex{
public:
	double* data;
	singleindex(int sizea){
		data = new double[sizea];
		memset(data,'\0',sizeof(double)*sizea);
	}
	~singleindex(){
		
	}
	
	double read(int a){return(data[a]);}
	void write(int a, double in){data[a] = in;}
	double& operator()(int a){
		return(data[a]);	
	}
	
	void clear(){delete[](data);}
};

int main(int argc, char * argv[]) {
	cout.precision(20);
	int END = 2;	
	
	
	string e_file;
	string inputs ("-i");
	
	double expected_l;
	string els ("-e");
	
	string total_lengths ("-l");
	int total_length;
	
	for(int i = 1;i < argc;++i){
		if(argv[i] == inputs){
			e_file = argv[i+1];
		}
		if(argv[i] == els){
			expected_l = atof(argv[i+1]);
		}
		if(argv[i] == total_lengths){
			total_length = atoi(argv[i+1]);
		}
	}
	
	
	double * e;
	e = new double[2 * total_length];
	char *p;
	
	
	ifstream file;
	file.open(e_file.c_str());
	if(!file) {
		cout << "Could not open the emission file." << endl;	
		return 0;
		
	}
	string line;
	int count_line = 0;
	while(getline(file,line)){
		if(line != ""){
			int found = line.find("	");
			if(found != string::npos){
				e[count_line*2] = strtod(line.substr(0,found).c_str(),&p);
				e[count_line*2+1] = strtod(line.substr(found+1).c_str(),&p);				
			}
			++count_line;
		}
	}
	file.close();
	
	double a[2][2];
	
	a[0][1] = 0.005;
	a[0][0] = 1-a[0][1];
	
	a[1][0] = 1/expected_l;
	a[1][1] = 1 - a[1][0];
	
	double begin[2];
	for(int i = 0;i < 2;++i){
		begin[i] = a[0][i];
	}
	int current_state = 1;
	int length_protein = total_length;
	double previous_P_x = -10000000;
	for(int iteration = 0;iteration < 1000;++iteration){
		
		doubleindex f = doubleindex(current_state+1,length_protein+1);	
		doubleindex b = doubleindex(current_state+1,length_protein+1);
		
		singleindex log_fs = singleindex(length_protein+1);
		singleindex log_bs = singleindex(length_protein+1);
		singleindex f_scaling = singleindex(length_protein+1);
		// Initiation of Forward Algorithm
		for(int i = 0;i<=current_state;++i){
			//f[i][0] = 1;	
			f(i,0) = 1;	
		}
		// Process initiation;
		double log_fs_process = 0;
		double log_bs_process = 0;
		
		//Processing of Viterbi and Forward algorithm : i = amino acids, 1 index.
		
		for(int i = 1;i<=length_protein;++i){
			//Go through all the states : j = states. Browse horizontally.
			double prev_f = 0;
			for(int j = 0;j <= current_state;++j){
				prev_f = 0;
				if(i == 1){
					f(j,i) = begin[j];
				}
				else{
					for(int k = 0;k<=current_state;++k){
								
						prev_f += (f(k,i-1) * a[k][j]);
					}
					f(j,i) = prev_f;
					
				}
				f(j,i) *=  exp(e[j + (current_state+1) * (i-1)]);
				f_scaling(i) += f(j,i);
		
			}
			
			for(int j = 0;j<=current_state;++j){
				f(j,i) = (f(j,i) / f_scaling(i));
			}
			//Scaling value log sum.
			log_fs_process += log(f_scaling(i));
			log_fs(i) = log_fs_process;	
		}
		//Forward termination
		double P_x = 0;
		for(int k = 0;k<=current_state;++k){
			int term;
			if(END == 2){
				if(k == 0 || k == current_state){
					term = 1;	
				}
				else{
					term = 0;	
				}
			}
			if(term == 1){
				P_x += f(k,length_protein);	
			}
		}
		P_x = log(P_x) + log_fs(length_protein);
		
		// Initiation of Backward Algorithm [scaled]	
		for(int k=0;k<=current_state;++k){
			int backward_init = 1;
			b(k,length_protein) = backward_init/f_scaling(length_protein);
		}
		
		//Implementation of Backward Algorithm
		for(int i = length_protein-1;i>=1;--i){
			//string prev_letter = gen.getSequenceByID(p).getSequence().substr(i,1);
			// Scaling values of b
			log_bs_process += log(f_scaling(i+1));
			log_bs(i+1) = log_bs_process;
			
			for(int k = 0;k <= current_state;++k){
				for(int j = 0;j<=current_state;++j){
					b(k,i) += (b(j,i+1) *  exp(e[j + (current_state+1) * i]) * a[k][j])/f_scaling(i);
				}
			}
		}
		//Termination of scaling values
		log_bs_process += log(f_scaling(1));
		log_bs(1) = log_bs_process;
		
		//Posterior
		string posterior_string;
		double posterior[length_protein+1];
		memset(posterior,'\0',sizeof(int)*(length_protein+1));
		
		double sum_a[4] = {0};
		// Posterior in Log Space
		for(int i = 1;i <= length_protein;++i){
			int posterior_state = 0;
			
			posterior[i] = 0;
			//#######################################################
			if(f(posterior_state,i) == 0 || b(posterior_state,i) == 0){
				posterior[i] = exp(-1 * pow(10,16));	
			}
			else{
				posterior[i] = exp(log(f(posterior_state,i)) + log_fs(i) + log(b(posterior_state,i)) + log_bs(i) - P_x);
			}
			if(i < length_protein){
				sum_a[0] += exp(log(f(0,i)) + log_fs(i) + log(b(0,i+1)) + log_bs(i+1) - P_x + log(a[0][0]) + e[0 + (current_state+1) * (i)]); // 0 to 0
				sum_a[1] += exp(log(f(0,i)) + log_fs(i) + log(b(1,i+1)) + log_bs(i+1) - P_x + log(a[0][1]) + e[1 + (current_state+1) * (i)]); // 0 to 1
				
				
				sum_a[2] += exp(log(f(1,i)) + log_fs(i) + log(b(0,i+1)) + log_bs(i+1) - P_x + log(a[1][0]) + e[0 + (current_state+1) * (i)]); // 1 to 0
				sum_a[3] += exp(log(f(1,i)) + log_fs(i) + log(b(1,i+1)) + log_bs(i+1) - P_x + log(a[1][1]) + e[1 + (current_state+1) * (i)]); // 1 to 1
			}
		}
		//cout << P_x << "	" << previous_P_x << "	" <<  1/a[1][0] << "	" << a[0][1] * length_protein << endl;
		//cout << a[0][0] << "	" << sum_a[0]/(sum_a[0] + sum_a[1]) << endl;
		//cout << a[0][1] << "	" << sum_a[1]/(sum_a[0] + sum_a[1]) << endl;
		//cout << a[1][0] << "	" << sum_a[2]/(sum_a[2] + sum_a[3]) << endl;
		//cout << a[1][1] << "	" << sum_a[3]/(sum_a[2] + sum_a[3]) << endl;
		
		f.clear();
		b.clear();
		
		log_fs.clear();
		log_bs.clear();
		f_scaling.clear();
		
		if(P_x < previous_P_x){
			break;	
		}
		
		a[0][0] = sum_a[0]/(sum_a[0] + sum_a[1]);
		a[0][1] = 1 - a[0][0];
		
		a[1][1] = sum_a[3]/(sum_a[2] + sum_a[3]);
		a[1][0] = 1 - a[1][1];
		
		if(fabs(previous_P_x - P_x) < 0.000001){
			break;	
		}
		else{
			previous_P_x = P_x;	
		}
		
	}
	cout << "Expected count : " << a[0][1] * length_protein << endl;
	cout << "Expected length : " << 1/a[1][0] << endl; 
}
	
	
