// ##############################################################################
// Computes the optimal local rate of evolution. Not for distribution.
// Copyright Alex Nguyen Ba 2011
// Version 1.1
//
// LICENCE 
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ##############################################################################


using namespace std;

#include <boost/regex.hpp>

#include <string>
#include <iostream>
#include <fstream>
#include <map>
#include <stdlib.h> 
#include <cstring>
#include <math.h>
#include <sstream>
#include <vector>

double P[26][26] = {0};
double P1[26][26] = {0};
double *lambda = new double[26];

double P_gap[2][2] = {0};
double P1_gap[2][2] = {0};
double *lambda_gap = new double[2];

//###############################################################################
// Symbols (Amino acids) Change to only ATCG if DNA will dramatically improve
// speed.
//###############################################################################

string aalist[20] = {"L","S","K","I","E","N","T","D","V","A","G","R","F","P","Q","Y","H","M","C","W"};
int aalist_size = sizeof(aalist)/sizeof(string);

string aalist_gap[2] = {"X","-"};
int aalist_size_gap = sizeof(aalist_gap)/sizeof(string);


class Sequence {
	
private:
	string m_seq_ID;
	string m_sequence;
public:
	
	//Constructor

	//Sequence(){
	//	m_seq_ID = NULL;
	//	m_sequence = NULL;
	//}
	
	//Setters
	bool appendLine(string seq_line){
		m_sequence += seq_line;
	}
	
	bool setHeader(string header){
		m_seq_ID = header;	
	}
	
	//Getters
	string getHeader(){
		return m_seq_ID;	
	}
	
	string getSequence(){
		return m_sequence;	
	}
};

class Genome {
private:
	int m_genome_size; //Number of sequences	
	map<int , Sequence*> list;
public:
	// Constructor
	Genome(){
		m_genome_size = -1;	
	}
	
	// Getter
	int getGenomeSize(){
		return m_genome_size+1;	
	}
	
	// Main function used.
	bool read_fasta(string fastafile){
		ifstream file;
		file.open(fastafile.c_str());
		int i =0;
		if (!file) return 0;
		string gene_name;
		string line;
		
		while( getline( file, line ) ){
			if (line[0] == '>') {
				++m_genome_size;
				string header;
				header = "";
				for(int j = 1;j < line.length();j++){
					header += line[j];	
				}
				list[m_genome_size] = new Sequence;
				list[m_genome_size]->setHeader(header);
			}
			else{
				if(line.rfind("*")!=string::npos){
					line = line.erase(line.rfind("*"),1);	
				}
				if(line.rfind("\n")!=string::npos){
					line = line.erase(line.rfind("\n"),1);	
				}
				if(line.rfind("\r")!=string::npos){
					line = line.erase(line.rfind("\r"),1);	
				}
				
				list[m_genome_size]->appendLine(line);
			}
		}		
		return 1;	
	}
	
	// Grab a child sequence.
	Sequence getSequenceByID(int ID){
		return *list[ID];	
	}
	//Delete a child sequence
	void deleteSequenceByID(int ID){
		for(int i = ID;i < m_genome_size;++i){
			//delete[](list[i]);
			list[i] = list[i+1];
		}
		//delete[](list[m_genome_size]);
		list.erase(m_genome_size);
		m_genome_size -= 1;
	}
};

class doubleindex{
public:
	int incsize;
	double* data;
	doubleindex(int sizea, int sizeb){
		incsize = sizea;
		data = new double[sizea*sizeb];
		memset(data,'\0',sizeof(double)*sizea*sizeb);
	}
	~doubleindex(){
	}
	
	double read(int a, int b){return(data[a + b * incsize]);}
	void write(int a, int b, double in ){data[a + b * incsize] =in ;}
	double& operator()(int a, int b){
		return(data[a + b * incsize]);
	}
	
	void clear(){delete[](data);}

};

class Node{
private:
	string m_sequence;
	string m_node_ID;
	bool is_parent;
	int child_1;
	int child_2;
	double distance_1;
	double distance_2;
public:
	Node(){
		is_parent = 0;	
	}
	// Getters
	bool isParent(){
		return is_parent;	
	}
	int getChild_1(){
		return child_1;	
	}
	int getChild_2(){
		return child_2;	
	}
	double getDistance_1(){
		return distance_1;	
	}
	double getDistance_2(){
		return distance_2;	
	}
	// Setters
	bool appendLine(string seq_line){
		m_sequence += seq_line;
	}
	bool set_parent(){
		is_parent = 1;
	}
	bool set_child_1(int childID){
		child_1 = childID;	
	}
	bool set_child_2(int childID){
		child_2 = childID;	
	}
	bool set_distance_1(double d){
		distance_1 = d;	
	}
	bool set_distance_2(double d){
		distance_2 = d;	
	}
	string getSequence(){
		return m_sequence;	
	}	
};
class NodeList{
private:
	int Nodelist_size;
	vector<Node*> list;
public:
	// Constructor
	NodeList(){
		Nodelist_size = -1;
	}
	
	bool createNode(){
		++Nodelist_size;
		list[Nodelist_size] = new Node;
	}
	bool resize_vector(int a){
		list.resize(2 * a); 
	}
	// Grab a node.
	Node getNodeByID(int ID){
		return *list[ID];	
	}
	
	// Getter
	int getNodeListSize(){
		return Nodelist_size+1;	
	}
	
	//
	void setSequenceOfNode(int ID,string seq){
		list[ID]->appendLine(seq);
	}
	
	//
	void setChildAsParent(int ID){
		list[ID]->set_parent();	
	}
	
	void setChild1(int ID,int childID){
		list[ID]->set_child_1(childID);	
	}
	void setChild2(int ID,int childID){
		list[ID]->set_child_2(childID);	
	}
	void setDistance1(int ID,double d){
		list[ID]->set_distance_1(d);	
	}
	void setDistance2(int ID,double d){
		list[ID]->set_distance_2(d);	
	}
};
int count_in_string(string haystack, string needle){
	int count_in_string = 0;
	int i = 0;
	while(haystack.find(needle,i)!=string::npos){
		i = haystack.find(needle,i)+1;
		++count_in_string;
	}
	return count_in_string;
}
string help(){
	string helpstr;
	helpstr = "You are using local rate estimator v.1.1 copyright Alex Nguyen Ba 2009\n";
	helpstr += "-i input file\n";
	helpstr += "-start [optional] Specify the position to start the algorithm in the gene.\n";
	helpstr += "-stop [optional] Specify the position to stop the algorithm in the gene.\n";
	return(helpstr);	
}

string convert(int number){
	ostringstream sin;
	sin << number;
	string val = sin.str();
	return(val);
}

int main(int argc, char * argv[]) {
	
	// ##################################################
	// Usage
	// ##################################################
	
	cout.precision(20);
	string fastafile;
	int start = 1;
	int stop = 0;
	double d_scale = 1;
	int window_size = 21;
	string treefile;
	int wag = 0;
	if(argc<2){
		cout << "usage: -i fastafile ... or try -help for more information." << endl;
		return 0;
	}
	string inputs ("-i");
	string starts ("-start");
	string stops ("-stop");
	string distances ("-distance");
	string gaps ("-gap");
	string windows ("-w");
	string trees ("-tree");
	string wags ("-wag");
	
	for(int i = 1;i < argc;++i){
		if(argv[i] == inputs){
			fastafile = argv[i+1];
		}
		if(argv[i] == starts){
			start = atoi(argv[i+1]);
		}
		if(argv[i] == stops){
			stop = atoi(argv[i+1]);
		}
		if(argv[i] == distances){
			d_scale = atof(argv[i+1]);
		}
		if(argv[i] == trees){
			treefile = argv[i+1];	
		}
		if(argv[i] == windows){
			window_size = atoi(argv[i+1]);
		}
		if(argv[i] == wags){
			wag = 1;
		}
	}
	
	if(fastafile.length() == 0){ 
		cout << "Requires a fasta file." << endl;
	}
	
	//###############################################################################
	// Handle the input fasta file.
	//###############################################################################
	Genome gen;
	if(!gen.read_fasta(fastafile)){
		cout << "Cannot read the fasta file." << endl;	
		return 0;
	}
	
	//###############################################################################
	// Sequence mapping
	// The code must recognize the following:
	// For a specie in the tree, what is the corresponding label in the fasta file.
	// By default, the mapping is the specie in the fasta file to the tree file.
	//###############################################################################
	int number_of_species = gen.getGenomeSize();
	
	string species_tree[number_of_species];
	string species_fasta[number_of_species];
	
	for(int i = 0;i < number_of_species;++i){
		string header = gen.getSequenceByID(i).getHeader();
		species_fasta[i] = header;
		species_tree[i] = header;
	}
	
	//###############################################################################
	// Phylogenetic tree in Newick format. 
	// Note, species not in this tree will not be parsed.
	// The tree will accomodate itself given the input file.
	// If the tree format is wrong, the code will likely fail.
	//###############################################################################
	
	ifstream file;
	file.open(treefile.c_str(),ifstream::in);
	if (!file){
		cout << "Can't find tree file." << endl;
		return 0;
	}
	string line = "";
	string tree = "";
	
	while(file >> line){
		tree += line;	
	}
	
	// Remove semi-colon.	
	if(tree.rfind(";")!=string::npos){
		tree = tree.erase(tree.rfind(";"),1);	
	}

	string print_tree = tree;
	
	//###############################################################################
	// Tree parsing
	// First step escapes all the tree_tags to regex format
	// Second step deletes empty sequences
	// Third step fills the node child array.
	// Fourth step removes unused species.
	// Last step fills parent array.
	//###############################################################################
	
	// Escape the tree string.
	for(int i = 0;i < number_of_species;++i){
		string::size_type pos = 0;
		if ( (pos = species_tree[i].find(".", pos)) != string::npos ) {
			species_tree[i].replace( pos, 1, "\\." );
			pos++;
		}
		pos = 0;
		if ( (pos = species_fasta[i].find(".", pos)) != string::npos ) {
			species_fasta[i].replace( pos, 1, "\\." );
			pos++;
		}
	}
	
	// Remove empty sequences.
	// Remove sequences not in the tree
	boost::regex pattern;
	for(int j = 0; j < gen.getGenomeSize();++j){
		pattern = gen.getSequenceByID(j).getHeader();
		if(!(regex_search(tree,pattern))){
			gen.deleteSequenceByID(j);
			--j;
			continue;
		}


		string my_sequence = gen.getSequenceByID(j).getSequence();
		int length_protein = my_sequence.length();
		
		if(length_protein == 0){
			gen.deleteSequenceByID(j);
			--j;
		}
	}
	
	
	boost::regex pattern2;
	std::string::const_iterator start_tree, end_tree;
	std::string::const_iterator start_line, end_line;
	boost::match_results<std::string::const_iterator> what; 
	boost::match_results<std::string::const_iterator> what2; 
	
	NodeList nodes;	
	nodes.resize_vector(number_of_species);
	// Populate the tree leaves into node IDs.

	for(int j = 0;j<gen.getGenomeSize();++j){
		for(int i=0;i<number_of_species;++i){
			
			pattern = species_fasta[i];
			if(regex_search(gen.getSequenceByID(j).getHeader(),pattern)){
				pattern2 = species_tree[i];
				stringstream out;
				out << j;
				tree = regex_replace(tree,pattern2,out.str());
				out.clear();
			}
		}
		nodes.createNode();
		
		nodes.setSequenceOfNode(nodes.getNodeListSize()-1,gen.getSequenceByID(j).getSequence());
	}
	
	
	start_tree = tree.begin(); 
	end_tree = tree.end();
	
	// Remove empty clades.
	pattern = "(.*)\\([A-Za-z\\.]+[0-9]*:[0-9\\.]+,[A-Za-z\\.]+[0-9]*:[0-9\\.]+\\):([0-9\\.]+)(.*)";
	while(regex_search(tree,what,pattern)){
		tree = what[1] + "REMOVE" + ":" + what[2] + what[3];
	}
	
	// Remove empty leaves.
	// Pattern 1 : ,REMOVE:0.1234)
	// Remove the pattern and a balanced parenthesis.
	start_tree = tree.begin(); 
	end_tree = tree.end();
	pattern = "(.*),[A-Za-z\\.]+[0-9]*:[0-9\\.]+\\)(.*)";
	while(regex_search(start_tree, end_tree, what, pattern)){
		// First part of the tree
		string first_part = what[1];
		// Last part of the tree
		string last_part = what[2];
		
		// We must remove a balanced parenthesis in the first part. 
		// Scan from right to left.
		int balanced = 0;
		for(int i = first_part.length()-1;i >= 0;--i){                                                     
			string letter = first_part.substr(i,1);
			
			if(letter == "("){
				if(balanced == 0){
					// Remove the encountered letter.
					tree = first_part.substr(0,i) + first_part.substr(i+1) + last_part;
					break;
				}
				--balanced;
			}
			else if(letter == ")"){
				++balanced;	
			}
		}
		pattern2 = "(.+:)([0-9\\.]+):([0-9\\.]+)(.+)";
		std::string::const_iterator start_tree2, end_tree2;
		
		start_tree2 = tree.begin();
		end_tree2 = tree.end();
		
		if(regex_search(start_tree2,end_tree2,what2,pattern2)){
			string first;
			string second;
			first = what2[2];
			second = what2[3];
			double sum = atof(first.c_str()) + atof(second.c_str());
			
			stringstream out;
			out << sum;
			tree = what2[1] + out.str() + what2[4];
			out.clear();
			
		}
		start_tree = tree.begin(); 
		end_tree = tree.end();
	}
	
	// Pattern 2 : (REMOVE:0.1234,
	// Remove the pattern and a balanced parenthesis.
	start_tree = tree.begin(); 
	end_tree = tree.end();
	pattern = "(.*)\\([A-Za-z\\.]+[0-9]*:[0-9\\.]+,(.*)";
	while(regex_search(start_tree, end_tree, what, pattern)){
		// First part of the tree
		string first_part = what[1];
		// Last part of the tree
		string last_part = what[2];
		
		// We must remove a balanced parenthesis in the last part. 
		// Scan from left to right
		int balanced = 0;
		for(int i = 0;i < last_part.length();++i){                                                     
			string letter = last_part.substr(i,1);
			
			if(letter == "("){
				++balanced;
				
			}
			else if(letter == ")"){
				if(balanced == 0){
					// Remove the encountered letter.
					tree = first_part + last_part.substr(0,i) + last_part.substr(i+1);
					break;
				}
				--balanced;	
			}
		}
		pattern2 = "(.+:)([0-9\\.]+):([0-9\\.]+)(.+)";
		std::string::const_iterator start_tree2, end_tree2;
		
		start_tree2 = tree.begin();
		end_tree2 = tree.end();
		
		if(regex_search(start_tree2,end_tree2,what2,pattern2)){
			string first;
			string second;
			first = what2[2];
			second = what2[3];
			double sum = atof(first.c_str()) + atof(second.c_str());
			
			stringstream out;
			out << sum;
			tree = what2[1] + out.str() + what2[4];
			out.clear();
			
		}
		start_tree = tree.begin(); 
		end_tree = tree.end();
	}
	
	start_tree = tree.begin(); 
	end_tree = tree.end();
	
	pattern = "(.*?):[0-9\\.]+$";
	if(regex_search(start_tree,end_tree,what,pattern)){
		tree = what[1];
	}
	
	int parent = gen.getGenomeSize();
	int children_size = parent;
	start_tree = tree.begin(); 
	end_tree = tree.end();
	pattern = "\\(([0-9]*):([0-9\\.]*),([0-9]*):([0-9\\.]*)\\)";

	while(regex_search(start_tree, end_tree, what, pattern)) { 
		
		string child_1 = what[1];
		string child_2 = what[3];
		string distance_1 = what[2];
		string distance_2 = what[4];
		
		nodes.createNode();
		int node_ID = nodes.getNodeListSize()-1;
		nodes.setChildAsParent(node_ID);
	
		nodes.setChild1(node_ID,atoi(child_1.c_str()));
		nodes.setChild2(node_ID,atoi(child_2.c_str()));
		
		nodes.setDistance1(node_ID,d_scale * atof(distance_1.c_str()));
		nodes.setDistance2(node_ID,d_scale * atof(distance_2.c_str()));
		
		pattern2 = "(.*?)\\(([0-9:\\.]*),([0-9:\\.]*)\\)(.*)";
		stringstream out;
		out << parent;
		//tree = regex_replace(tree,pattern2,out.str());
		if(regex_search(start_tree,end_tree,what2,pattern2)){

			tree = what2[1] + out.str() + what2[4];	
		}
		out.clear();
		++parent;
				
		start_tree = tree.begin();
		end_tree = tree.end();
	}
	if(parent == 0){
		cout << "Failed to parse the tree." << endl;
		return 0;
	}
	// We should make sure all child are equal in length; Skip this for now.
	if(stop == 0){ 
		stop = gen.getSequenceByID(0).getSequence().length();
	}	
	
	// Populate substitution matrix
	// We will hardcode the matrix instead. Lambdas in log space.
	double * sum_letters;
	sum_letters = new double[26];
	if(wag == 0){
		lambda[(int)('L' - 'A')]  = -0.262352069541849;
		lambda[(int)('S' - 'A')]  = -0.254867733800273;
		lambda[(int)('K' - 'A')]  = -0.225853382091412;
		lambda[(int)('I' - 'A')]  = -0.212332452414312;
		lambda[(int)('E' - 'A')]  = -0.188558928817862;
		lambda[(int)('N' - 'A')]  = -0.167140155639894;
		lambda[(int)('T' - 'A')]  = -0.14395950834156;
		lambda[(int)('D' - 'A')]  = -0.126531662282891;
		lambda[(int)('V' - 'A')]  = -0.124281165343963;
		lambda[(int)('A' - 'A')]  = -0.118051382908287;
		lambda[(int)('G' - 'A')]  = -0.112887472255798;
		lambda[(int)('R' - 'A')]  = -0.106768506411763;
		lambda[(int)('F' - 'A')]  = -0.0804401241329973;
		lambda[(int)('P' - 'A')]  = -0.0768501808813535;
		lambda[(int)('Q' - 'A')]  = -0.0748240469356219;
		lambda[(int)('Y' - 'A')]  = -0.0555950811873711;
		lambda[(int)('H' - 'A')]  = -0.0423826387755612;
		lambda[(int)('M' - 'A')]  = -0.0392766891199065;
		lambda[(int)('C' - 'A')]  = -0.0100967308691275;
		lambda[(int)('W' - 'A')]  = 0;
	
		P[(int)('L' - 'A')][(int)('L' - 'A')] = -0.049681237;	P[(int)('L' - 'A')][(int)('S' - 'A')] = -0.002415121;	P[(int)('L' - 'A')][(int)('K' - 'A')] = -0.017779188;	P[(int)('L' - 'A')][(int)('I' - 'A')] = -0.026492808;	P[(int)('L' - 'A')][(int)('E' - 'A')] = 0.003190431;	P[(int)('L' - 'A')][(int)('N' - 'A')] = -0.065180382;	P[(int)('L' - 'A')][(int)('T' - 'A')] = 0.22015554;	P[(int)('L' - 'A')][(int)('D' - 'A')] = 0.73416107;	P[(int)('L' - 'A')][(int)('V' - 'A')] = 0.29055798;	P[(int)('L' - 'A')][(int)('A' - 'A')] = -0.18070275;	P[(int)('L' - 'A')][(int)('G' - 'A')] = -0.12485798;	P[(int)('L' - 'A')][(int)('R' - 'A')] = 0.28014971;	P[(int)('L' - 'A')][(int)('F' - 'A')] = -0.1813374;	P[(int)('L' - 'A')][(int)('P' - 'A')] = -0.012401368;	P[(int)('L' - 'A')][(int)('Q' - 'A')] = -0.046902305;	P[(int)('L' - 'A')][(int)('Y' - 'A')] = -0.035150013;	P[(int)('L' - 'A')][(int)('H' - 'A')] = -0.32385684;	P[(int)('L' - 'A')][(int)('M' - 'A')] = 0.03378741;	P[(int)('L' - 'A')][(int)('C' - 'A')] = -0.033768978;	P[(int)('L' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('S' - 'A')][(int)('L' - 'A')] = 0.81716129;	P[(int)('S' - 'A')][(int)('S' - 'A')] = 0.061249572;	P[(int)('S' - 'A')][(int)('K' - 'A')] = 0.15587172;	P[(int)('S' - 'A')][(int)('I' - 'A')] = 0.31333291;	P[(int)('S' - 'A')][(int)('E' - 'A')] = 0.066562022;	P[(int)('S' - 'A')][(int)('N' - 'A')] = 0.15059126;	P[(int)('S' - 'A')][(int)('T' - 'A')] = 0.23673576;	P[(int)('S' - 'A')][(int)('D' - 'A')] = 0.040072747;	P[(int)('S' - 'A')][(int)('V' - 'A')] = -0.10097594;	P[(int)('S' - 'A')][(int)('A' - 'A')] = -0.031716243;	P[(int)('S' - 'A')][(int)('G' - 'A')] = 0.032901684;	P[(int)('S' - 'A')][(int)('R' - 'A')] = -0.070946219;	P[(int)('S' - 'A')][(int)('F' - 'A')] = 0.065944418;	P[(int)('S' - 'A')][(int)('P' - 'A')] = -0.036255272;	P[(int)('S' - 'A')][(int)('Q' - 'A')] = 0.19477143;	P[(int)('S' - 'A')][(int)('Y' - 'A')] = 0.079643091;	P[(int)('S' - 'A')][(int)('H' - 'A')] = -0.001441757;	P[(int)('S' - 'A')][(int)('M' - 'A')] = -0.037550607;	P[(int)('S' - 'A')][(int)('C' - 'A')] = -0.050915707;	P[(int)('S' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('K' - 'A')][(int)('L' - 'A')] = 0.10517411;	P[(int)('K' - 'A')][(int)('S' - 'A')] = 0.00579841;	P[(int)('K' - 'A')][(int)('K' - 'A')] = -0.15827075;	P[(int)('K' - 'A')][(int)('I' - 'A')] = -0.41563643;	P[(int)('K' - 'A')][(int)('E' - 'A')] = 0.59004463;	P[(int)('K' - 'A')][(int)('N' - 'A')] = 0.045540384;	P[(int)('K' - 'A')][(int)('T' - 'A')] = 0.002593839;	P[(int)('K' - 'A')][(int)('D' - 'A')] = 0.086828695;	P[(int)('K' - 'A')][(int)('V' - 'A')] = 0.026265838;	P[(int)('K' - 'A')][(int)('A' - 'A')] = 0.15779057;	P[(int)('K' - 'A')][(int)('G' - 'A')] = -0.20092364;	P[(int)('K' - 'A')][(int)('R' - 'A')] = -0.39370815;	P[(int)('K' - 'A')][(int)('F' - 'A')] = -0.24896154;	P[(int)('K' - 'A')][(int)('P' - 'A')] = 0.028638242;	P[(int)('K' - 'A')][(int)('Q' - 'A')] = -0.15548169;	P[(int)('K' - 'A')][(int)('Y' - 'A')] = -0.07368193;	P[(int)('K' - 'A')][(int)('H' - 'A')] = 0.20682824;	P[(int)('K' - 'A')][(int)('M' - 'A')] = -0.15559485;	P[(int)('K' - 'A')][(int)('C' - 'A')] = -0.054511853;	P[(int)('K' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('I' - 'A')][(int)('L' - 'A')] = 0.058081544;	P[(int)('I' - 'A')][(int)('S' - 'A')] = -0.59299062;	P[(int)('I' - 'A')][(int)('K' - 'A')] = -0.25514888;	P[(int)('I' - 'A')][(int)('I' - 'A')] = 0.19477615;	P[(int)('I' - 'A')][(int)('E' - 'A')] = 0.063859591;	P[(int)('I' - 'A')][(int)('N' - 'A')] = 0.18749096;	P[(int)('I' - 'A')][(int)('T' - 'A')] = -0.41081291;	P[(int)('I' - 'A')][(int)('D' - 'A')] = -0.00201392;	P[(int)('I' - 'A')][(int)('V' - 'A')] = 0.11120814;	P[(int)('I' - 'A')][(int)('A' - 'A')] = 0.12305416;	P[(int)('I' - 'A')][(int)('G' - 'A')] = 0.28544197;	P[(int)('I' - 'A')][(int)('R' - 'A')] = -0.071589204;	P[(int)('I' - 'A')][(int)('F' - 'A')] = 0.076413924;	P[(int)('I' - 'A')][(int)('P' - 'A')] = -0.017739701;	P[(int)('I' - 'A')][(int)('Q' - 'A')] = -0.17796799;	P[(int)('I' - 'A')][(int)('Y' - 'A')] = 0.067092985;	P[(int)('I' - 'A')][(int)('H' - 'A')] = -0.34091978;	P[(int)('I' - 'A')][(int)('M' - 'A')] = -0.089891867;	P[(int)('I' - 'A')][(int)('C' - 'A')] = -0.056580404;	P[(int)('I' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('E' - 'A')][(int)('L' - 'A')] = -0.047533081;	P[(int)('E' - 'A')][(int)('S' - 'A')] = -0.014890516;	P[(int)('E' - 'A')][(int)('K' - 'A')] = 0.15818656;	P[(int)('E' - 'A')][(int)('I' - 'A')] = 0.38193051;	P[(int)('E' - 'A')][(int)('E' - 'A')] = 0.19960266;	P[(int)('E' - 'A')][(int)('N' - 'A')] = -0.56531243;	P[(int)('E' - 'A')][(int)('T' - 'A')] = -0.12978641;	P[(int)('E' - 'A')][(int)('D' - 'A')] = -0.044030883;	P[(int)('E' - 'A')][(int)('V' - 'A')] = 0.15943821;	P[(int)('E' - 'A')][(int)('A' - 'A')] = 0.35145129;	P[(int)('E' - 'A')][(int)('G' - 'A')] = -0.12259817;	P[(int)('E' - 'A')][(int)('R' - 'A')] = 0.26434271;	P[(int)('E' - 'A')][(int)('F' - 'A')] = 0.070585802;	P[(int)('E' - 'A')][(int)('P' - 'A')] = 0.26961129;	P[(int)('E' - 'A')][(int)('Q' - 'A')] = -0.000086778831;	P[(int)('E' - 'A')][(int)('Y' - 'A')] = -0.05864428;	P[(int)('E' - 'A')][(int)('H' - 'A')] = 0.21638688;	P[(int)('E' - 'A')][(int)('M' - 'A')] = -0.18148197;	P[(int)('E' - 'A')][(int)('C' - 'A')] = -0.060874437;	P[(int)('E' - 'A')][(int)('W' - 'A')] = 0.2236068;	P[(int)('N' - 'A')][(int)('L' - 'A')] = -0.44696064;
		P[(int)('N' - 'A')][(int)('S' - 'A')] = -0.074629599;	P[(int)('N' - 'A')][(int)('K' - 'A')] = 0.40818135;	P[(int)('N' - 'A')][(int)('I' - 'A')] = 0.22989943;	P[(int)('N' - 'A')][(int)('E' - 'A')] = 0.1947956;	P[(int)('N' - 'A')][(int)('N' - 'A')] = 0.53009713;	P[(int)('N' - 'A')][(int)('T' - 'A')] = 0.15888675;	P[(int)('N' - 'A')][(int)('D' - 'A')] = 0.15108131;	P[(int)('N' - 'A')][(int)('V' - 'A')] = -0.30671582;	P[(int)('N' - 'A')][(int)('A' - 'A')] = 0.070081118;	P[(int)('N' - 'A')][(int)('G' - 'A')] = 0.026053068;	P[(int)('N' - 'A')][(int)('R' - 'A')] = 0.013175221;	P[(int)('N' - 'A')][(int)('F' - 'A')] = 0.11931146;	P[(int)('N' - 'A')][(int)('P' - 'A')] = 0.1047864;	P[(int)('N' - 'A')][(int)('Q' - 'A')] = 0.048731598;	P[(int)('N' - 'A')][(int)('Y' - 'A')] = -0.011323853;	P[(int)('N' - 'A')][(int)('H' - 'A')] = 0.13459322;	P[(int)('N' - 'A')][(int)('M' - 'A')] = -0.095339243;	P[(int)('N' - 'A')][(int)('C' - 'A')] = -0.0583983;	P[(int)('N' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('T' - 'A')][(int)('L' - 'A')] = -0.25165371;	P[(int)('T' - 'A')][(int)('S' - 'A')] = 0.28614434;	P[(int)('T' - 'A')][(int)('K' - 'A')] = -0.61615283;	P[(int)('T' - 'A')][(int)('I' - 'A')] = 0.27315449;	P[(int)('T' - 'A')][(int)('E' - 'A')] = 0.007468631;	P[(int)('T' - 'A')][(int)('N' - 'A')] = -0.11026103;	P[(int)('T' - 'A')][(int)('T' - 'A')] = 0.43010446;	P[(int)('T' - 'A')][(int)('D' - 'A')] = -0.075518941;	P[(int)('T' - 'A')][(int)('V' - 'A')] = -0.09575316;	P[(int)('T' - 'A')][(int)('A' - 'A')] = 0.070162308;	P[(int)('T' - 'A')][(int)('G' - 'A')] = 0.19633815;	P[(int)('T' - 'A')][(int)('R' - 'A')] = -0.2000809;	P[(int)('T' - 'A')][(int)('F' - 'A')] = 0.11319097;	P[(int)('T' - 'A')][(int)('P' - 'A')] = -0.038537922;	P[(int)('T' - 'A')][(int)('Q' - 'A')] = 0.080977783;	P[(int)('T' - 'A')][(int)('Y' - 'A')] = 0.074447331;	P[(int)('T' - 'A')][(int)('H' - 'A')] = -0.11512003;	P[(int)('T' - 'A')][(int)('M' - 'A')] = -0.10933631;	P[(int)('T' - 'A')][(int)('C' - 'A')] = -0.057889781;	P[(int)('T' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('D' - 'A')][(int)('L' - 'A')] = 0.14269558;	P[(int)('D' - 'A')][(int)('S' - 'A')] = 0.028835636;	P[(int)('D' - 'A')][(int)('K' - 'A')] = -0.29242293;	P[(int)('D' - 'A')][(int)('I' - 'A')] = -0.38034106;	P[(int)('D' - 'A')][(int)('E' - 'A')] = -0.42496808;	P[(int)('D' - 'A')][(int)('N' - 'A')] = 0.19846483;	P[(int)('D' - 'A')][(int)('T' - 'A')] = -0.09024041;	P[(int)('D' - 'A')][(int)('D' - 'A')] = 0.14867582;	P[(int)('D' - 'A')][(int)('V' - 'A')] = -0.19645181;	P[(int)('D' - 'A')][(int)('A' - 'A')] = 0.2957769;	P[(int)('D' - 'A')][(int)('G' - 'A')] = -0.10008594;	P[(int)('D' - 'A')][(int)('R' - 'A')] = 0.28964442;	P[(int)('D' - 'A')][(int)('F' - 'A')] = 0.23053356;	P[(int)('D' - 'A')][(int)('P' - 'A')] = 0.31858099;	P[(int)('D' - 'A')][(int)('Q' - 'A')] = 0.070353596;	P[(int)('D' - 'A')][(int)('Y' - 'A')] = -0.030264684;	P[(int)('D' - 'A')][(int)('H' - 'A')] = 0.20105761;	P[(int)('D' - 'A')][(int)('M' - 'A')] = -0.15388422;	P[(int)('D' - 'A')][(int)('C' - 'A')] = -0.06167617;	P[(int)('D' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('V' - 'A')][(int)('L' - 'A')] = -0.00671921;	P[(int)('V' - 'A')][(int)('S' - 'A')] = 0.64479343;	P[(int)('V' - 'A')][(int)('K' - 'A')] = 0.29093514;	P[(int)('V' - 'A')][(int)('I' - 'A')] = -0.15579176;	P[(int)('V' - 'A')][(int)('E' - 'A')] = -0.015026826;	P[(int)('V' - 'A')][(int)('N' - 'A')] = 0.044523791;	P[(int)('V' - 'A')][(int)('T' - 'A')] = -0.34448917;	P[(int)('V' - 'A')][(int)('D' - 'A')] = -0.001989732;	P[(int)('V' - 'A')][(int)('V' - 'A')] = 0.150836;	P[(int)('V' - 'A')][(int)('A' - 'A')] = 0.1265477;	P[(int)('V' - 'A')][(int)('G' - 'A')] = 0.34671352;	P[(int)('V' - 'A')][(int)('R' - 'A')] = -0.098619711;	P[(int)('V' - 'A')][(int)('F' - 'A')] = 0.10119853;	P[(int)('V' - 'A')][(int)('P' - 'A')] = -0.006939004;	P[(int)('V' - 'A')][(int)('Q' - 'A')] = -0.11954331;	P[(int)('V' - 'A')][(int)('Y' - 'A')] = 0.070286963;	P[(int)('V' - 'A')][(int)('H' - 'A')] = -0.30566775;	P[(int)('V' - 'A')][(int)('M' - 'A')] = -0.097616607;	P[(int)('V' - 'A')][(int)('C' - 'A')] = -0.056684905;	P[(int)('V' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('A' - 'A')][(int)('L' - 'A')] = -0.056616309;	P[(int)('A' - 'A')][(int)('S' - 'A')] = -0.37343765;	P[(int)('A' - 'A')][(int)('K' - 'A')] = 0.34804428;	P[(int)('A' - 'A')][(int)('I' - 'A')] = -0.42188597;	P[(int)('A' - 'A')][(int)('E' - 'A')] = -0.16969608;	P[(int)('A' - 'A')][(int)('N' - 'A')] = -0.35211762;	P[(int)('A' - 'A')][(int)('T' - 'A')] = 0.39387541;	P[(int)('A' - 'A')][(int)('D' - 'A')] = -0.1017408;	P[(int)('A' - 'A')][(int)('V' - 'A')] = 0.010508699;	P[(int)('A' - 'A')][(int)('A' - 'A')] = 0.03547992;	P[(int)('A' - 'A')][(int)('G' - 'A')] = 0.26407098;	P[(int)('A' - 'A')][(int)('R' - 'A')] = -0.19780946;	P[(int)('A' - 'A')][(int)('F' - 'A')] = 0.15314603;	P[(int)('A' - 'A')][(int)('P' - 'A')] = -0.021309736;	P[(int)('A' - 'A')][(int)('Q' - 'A')] = 0.1374577;	P[(int)('A' - 'A')][(int)('Y' - 'A')] = 0.08232548;	P[(int)('A' - 'A')][(int)('H' - 'A')] = -0.11543601;	P[(int)('A' - 'A')][(int)('M' - 'A')] = -0.11560482;	P[(int)('A' - 'A')][(int)('C' - 'A')] = -0.058673043;	P[(int)('A' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('G' - 'A')][(int)('L' - 'A')] = -0.058954396;	P[(int)('G' - 'A')][(int)('S' - 'A')] = 0.010706691;	P[(int)('G' - 'A')][(int)('K' - 'A')] = -0.064696527;	P[(int)('G' - 'A')][(int)('I' - 'A')] = -0.021783595;	P[(int)('G' - 'A')][(int)('E' - 'A')] = 0.05523449;	P[(int)('G' - 'A')][(int)('N' - 'A')] = 0.027676821;	P[(int)('G' - 'A')][(int)('T' - 'A')] = -0.15996143;	P[(int)('G' - 'A')][(int)('D' - 'A')] = -0.13748663;	P[(int)('G' - 'A')][(int)('V' - 'A')] = 0.26397134;	P[(int)('G' - 'A')][(int)('A' - 'A')] = -0.7291697;	P[(int)('G' - 'A')][(int)('G' - 'A')] = -0.11893473;	P[(int)('G' - 'A')][(int)('R' - 'A')] = -0.092767701;	P[(int)('G' - 'A')][(int)('F' - 'A')] = 0.29056045;	P[(int)('G' - 'A')][(int)('P' - 'A')] = 0.33461851;	P[(int)('G' - 'A')][(int)('Q' - 'A')] = 0.16088686;	P[(int)('G' - 'A')][(int)('Y' - 'A')] = 0.053018954;	P[(int)('G' - 'A')][(int)('H' - 'A')] = 0.13845924;	P[(int)('G' - 'A')][(int)('M' - 'A')] = -0.14782752;	P[(int)('G' - 'A')][(int)('C' - 'A')] = -0.056634067;	P[(int)('G' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('R' - 'A')][(int)('L' - 'A')] = -0.052229604;	P[(int)('R' - 'A')][(int)('S' - 'A')] = -0.006448166;	P[(int)('R' - 'A')][(int)('K' - 'A')] = 0.097640225;	P[(int)('R' - 'A')][(int)('I' - 'A')] = 0.24537214;	P[(int)('R' - 'A')][(int)('E' - 'A')] = -0.57456348;	P[(int)('R' - 'A')][(int)('N' - 'A')] = -0.072144571;	P[(int)('R' - 'A')][(int)('T' - 'A')] = -0.13903561;	P[(int)('R' - 'A')][(int)('D' - 'A')] = 0.14098981;	P[(int)('R' - 'A')][(int)('V' - 'A')] = -0.024022526;	P[(int)('R' - 'A')][(int)('A' - 'A')] = -0.049717429;	P[(int)('R' - 'A')][(int)('G' - 'A')] = -0.16132325;	P[(int)('R' - 'A')][(int)('R' - 'A')] = -0.49792519;	P[(int)('R' - 'A')][(int)('F' - 'A')] = -0.33609857;	P[(int)('R' - 'A')][(int)('P' - 'A')] = -0.070869891;	P[(int)('R' - 'A')][(int)('Q' - 'A')] = -0.19068849;	P[(int)('R' - 'A')][(int)('Y' - 'A')] = -0.074658126;	P[(int)('R' - 'A')][(int)('H' - 'A')] = 0.22846803;	P[(int)('R' - 'A')][(int)('M' - 'A')] = -0.1090617;	P[(int)('R' - 'A')][(int)('C' - 'A')] = -0.043999086;	P[(int)('R' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('F' - 'A')][(int)('L' - 'A')] = -0.019895157;	P[(int)('F' - 'A')][(int)('S' - 'A')] = 0.001037422;	P[(int)('F' - 'A')][(int)('K' - 'A')] = -0.002052849;	P[(int)('F' - 'A')][(int)('I' - 'A')] = -0.010189739;	P[(int)('F' - 'A')][(int)('E' - 'A')] = -0.002269018;	P[(int)('F' - 'A')][(int)('N' - 'A')] = -0.00604177;	P[(int)('F' - 'A')][(int)('T' - 'A')] = -0.040261502;	P[(int)('F' - 'A')][(int)('D' - 'A')] = -0.20174552;	P[(int)('F' - 'A')][(int)('V' - 'A')] = -0.14090331;	P[(int)('F' - 'A')][(int)('A' - 'A')] = 0.011250757;	P[(int)('F' - 'A')][(int)('G' - 'A')] = 0.044979776;	P[(int)('F' - 'A')][(int)('R' - 'A')] = -0.046092584;	P[(int)('F' - 'A')][(int)('F' - 'A')] = -0.37647547;	P[(int)('F' - 'A')][(int)('P' - 'A')] = 0.45009843;	P[(int)('F' - 'A')][(int)('Q' - 'A')] = 0.29157229;	P[(int)('F' - 'A')][(int)('Y' - 'A')] = -0.3754672;	P[(int)('F' - 'A')][(int)('H' - 'A')] = -0.26779468;	P[(int)('F' - 'A')][(int)('M' - 'A')] = 0.48930639;	P[(int)('F' - 'A')][(int)('C' - 'A')] = -0.039446127;	P[(int)('F' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('P' - 'A')][(int)('L' - 'A')] = -0.12562282;	P[(int)('P' - 'A')][(int)('S' - 'A')] = -0.002587705;	P[(int)('P' - 'A')][(int)('K' - 'A')] = -0.030032023;	P[(int)('P' - 'A')][(int)('I' - 'A')] = -0.051698466;	P[(int)('P' - 'A')][(int)('E' - 'A')] = 0.01066163;	P[(int)('P' - 'A')][(int)('N' - 'A')] = -0.001360316;	P[(int)('P' - 'A')][(int)('T' - 'A')] = -0.27446871;	P[(int)('P' - 'A')][(int)('D' - 'A')] = -0.058605507;	P[(int)('P' - 'A')][(int)('V' - 'A')] = -0.00208106;	P[(int)('P' - 'A')][(int)('A' - 'A')] = 0.071722469;	P[(int)('P' - 'A')][(int)('G' - 'A')] = -0.26755618;	P[(int)('P' - 'A')][(int)('R' - 'A')] = 0.071077241;	P[(int)('P' - 'A')][(int)('F' - 'A')] = -0.081213172;	P[(int)('P' - 'A')][(int)('P' - 'A')] = -0.54047866;	P[(int)('P' - 'A')][(int)('Q' - 'A')] = 0.67396246;	P[(int)('P' - 'A')][(int)('Y' - 'A')] = 0.060779576;	P[(int)('P' - 'A')][(int)('H' - 'A')] = -0.057459336;	P[(int)('P' - 'A')][(int)('M' - 'A')] = -0.082476431;	P[(int)('P' - 'A')][(int)('C' - 'A')] = -0.05684524;	P[(int)('P' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('Q' - 'A')][(int)('L' - 'A')] = -0.000118842;	P[(int)('Q' - 'A')][(int)('S' - 'A')] = 0.000522634;	P[(int)('Q' - 'A')][(int)('K' - 'A')] = -0.010300904;	P[(int)('Q' - 'A')][(int)('I' - 'A')] = -0.03107875;	P[(int)('Q' - 'A')][(int)('E' - 'A')] = -0.089795219;	P[(int)('Q' - 'A')][(int)('N' - 'A')] = 0.27649482;	P[(int)('Q' - 'A')][(int)('T' - 'A')] = 0.22743903;	P[(int)('Q' - 'A')][(int)('D' - 'A')] = -0.38928252;	P[(int)('Q' - 'A')][(int)('V' - 'A')] = 0.44373145;	P[(int)('Q' - 'A')][(int)('A' - 'A')] = -0.01490388;	P[(int)('Q' - 'A')][(int)('G' - 'A')] = 0.21755774;	P[(int)('Q' - 'A')][(int)('R' - 'A')] = 0.36884631;	P[(int)('Q' - 'A')][(int)('F' - 'A')] = -0.35675943;	P[(int)('Q' - 'A')][(int)('P' - 'A')] = -0.17926459;	P[(int)('Q' - 'A')][(int)('Q' - 'A')] = -0.13555852;	P[(int)('Q' - 'A')][(int)('Y' - 'A')] = -0.15474384;	P[(int)('Q' - 'A')][(int)('H' - 'A')] = 0.25156734;	P[(int)('Q' - 'A')][(int)('M' - 'A')] = -0.11299245;	P[(int)('Q' - 'A')][(int)('C' - 'A')] = -0.053467688;	P[(int)('Q' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('Y' - 'A')][(int)('L' - 'A')] = -0.004703125;	P[(int)('Y' - 'A')][(int)('S' - 'A')] = -0.00014131;	P[(int)('Y' - 'A')][(int)('K' - 'A')] = -0.000049959719;	P[(int)('Y' - 'A')][(int)('I' - 'A')] = 0.002107193;	P[(int)('Y' - 'A')][(int)('E' - 'A')] = -0.015299426;	P[(int)('Y' - 'A')][(int)('N' - 'A')] = 0.028701239;	P[(int)('Y' - 'A')][(int)('T' - 'A')] = 0.041402345;	P[(int)('Y' - 'A')][(int)('D' - 'A')] = 0.004054481;	P[(int)('Y' - 'A')][(int)('V' - 'A')] = 0.22430225;	P[(int)('Y' - 'A')][(int)('A' - 'A')] = 0.10998942;	P[(int)('Y' - 'A')][(int)('G' - 'A')] = -0.16526757;	P[(int)('Y' - 'A')][(int)('R' - 'A')] = -0.11546305;	P[(int)('Y' - 'A')][(int)('F' - 'A')] = 0.52939247;	P[(int)('Y' - 'A')][(int)('P' - 'A')] = -0.28923863;	P[(int)('Y' - 'A')][(int)('Q' - 'A')] = -0.18324817;	P[(int)('Y' - 'A')][(int)('Y' - 'A')] = -0.34477172;	P[(int)('Y' - 'A')][(int)('H' - 'A')] = 0.13100243;	P[(int)('Y' - 'A')][(int)('M' - 'A')] = 0.55575385;	P[(int)('Y' - 'A')][(int)('C' - 'A')] = -0.041760599;	P[(int)('Y' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('H' - 'A')][(int)('L' - 'A')] = 0.038982232;	P[(int)('H' - 'A')][(int)('S' - 'A')] = 0.007239687;	P[(int)('H' - 'A')][(int)('K' - 'A')] = -0.060140373;	P[(int)('H' - 'A')][(int)('I' - 'A')] = -0.062028214;	P[(int)('H' - 'A')][(int)('E' - 'A')] = 0.11571894;	P[(int)('H' - 'A')][(int)('N' - 'A')] = -0.27078197;	P[(int)('H' - 'A')][(int)('T' - 'A')] = -0.18122503;	P[(int)('H' - 'A')][(int)('D' - 'A')] = 0.11423333;	P[(int)('H' - 'A')][(int)('V' - 'A')] = -0.52948374;	P[(int)('H' - 'A')][(int)('A' - 'A')] = -0.35321224;	P[(int)('H' - 'A')][(int)('G' - 'A')] = 0.36961101;	P[(int)('H' - 'A')][(int)('R' - 'A')] = 0.25098882;	P[(int)('H' - 'A')][(int)('F' - 'A')] = -0.053141185;	P[(int)('H' - 'A')][(int)('P' - 'A')] = -0.27033524;	P[(int)('H' - 'A')][(int)('Q' - 'A')] = -0.16669016;	P[(int)('H' - 'A')][(int)('Y' - 'A')] = -0.18630698;	P[(int)('H' - 'A')][(int)('H' - 'A')] = 0.24147237;	P[(int)('H' - 'A')][(int)('M' - 'A')] = 0.064615376;	P[(int)('H' - 'A')][(int)('C' - 'A')] = -0.052473005;	P[(int)('H' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('M' - 'A')][(int)('L' - 'A')] = 0.009115137;	P[(int)('M' - 'A')][(int)('S' - 'A')] = 0.025056105;	P[(int)('M' - 'A')][(int)('K' - 'A')] = 0.06398273;	P[(int)('M' - 'A')][(int)('I' - 'A')] = -0.029439419;	P[(int)('M' - 'A')][(int)('E' - 'A')] = -0.028172138;	P[(int)('M' - 'A')][(int)('N' - 'A')] = -0.027853075;	P[(int)('M' - 'A')][(int)('T' - 'A')] = 0.092631083;	P[(int)('M' - 'A')][(int)('D' - 'A')] = -0.37531455;	P[(int)('M' - 'A')][(int)('V' - 'A')] = -0.29037999;	P[(int)('M' - 'A')][(int)('A' - 'A')] = -0.10286728;	P[(int)('M' - 'A')][(int)('G' - 'A')] = -0.53120924;	P[(int)('M' - 'A')][(int)('R' - 'A')] = 0.18304029;	P[(int)('M' - 'A')][(int)('F' - 'A')] = 0.009323667;	P[(int)('M' - 'A')][(int)('P' - 'A')] = -0.080292469;	P[(int)('M' - 'A')][(int)('Q' - 'A')] = -0.41227551;	P[(int)('M' - 'A')][(int)('Y' - 'A')] = 0.079080378;	P[(int)('M' - 'A')][(int)('H' - 'A')] = -0.42820641;	P[(int)('M' - 'A')][(int)('M' - 'A')] = -0.11641769;	P[(int)('M' - 'A')][(int)('C' - 'A')] = -0.055281138;	P[(int)('M' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('C' - 'A')][(int)('L' - 'A')] = -0.048576489;	P[(int)('C' - 'A')][(int)('S' - 'A')] = -0.003272976;	P[(int)('C' - 'A')][(int)('K' - 'A')] = -0.015034294;	P[(int)('C' - 'A')][(int)('I' - 'A')] = -0.032018034;	P[(int)('C' - 'A')][(int)('E' - 'A')] = 0.008686522;	P[(int)('C' - 'A')][(int)('N' - 'A')] = -0.019220409;	P[(int)('C' - 'A')][(int)('T' - 'A')] = -0.030547168;	P[(int)('C' - 'A')][(int)('D' - 'A')] = -0.019603867;	P[(int)('C' - 'A')][(int)('V' - 'A')] = 0.012652466;	P[(int)('C' - 'A')][(int)('A' - 'A')] = 0.032529991;	P[(int)('C' - 'A')][(int)('G' - 'A')] = 0.00431469;	P[(int)('C' - 'A')][(int)('R' - 'A')] = 0.06268089;	P[(int)('C' - 'A')][(int)('F' - 'A')] = -0.14034989;	P[(int)('C' - 'A')][(int)('P' - 'A')] = 0.056782361;	P[(int)('C' - 'A')][(int)('Q' - 'A')] = -0.075644002;	P[(int)('C' - 'A')][(int)('Y' - 'A')] = 0.78980015;	P[(int)('C' - 'A')][(int)('H' - 'A')] = 0.2055114;	P[(int)('C' - 'A')][(int)('M' - 'A')] = 0.49418157;	P[(int)('C' - 'A')][(int)('C' - 'A')] = -0.023902416;	P[(int)('C' - 'A')][(int)('W' - 'A')] = 0.2236068;
		P[(int)('W' - 'A')][(int)('L' - 'A')] = -0.001945281;	P[(int)('W' - 'A')][(int)('S' - 'A')] = -0.000570264;	P[(int)('W' - 'A')][(int)('K' - 'A')] = -0.000760496;	P[(int)('W' - 'A')][(int)('I' - 'A')] = -0.002188577;	P[(int)('W' - 'A')][(int)('E' - 'A')] = 0.003965127;	P[(int)('W' - 'A')][(int)('N' - 'A')] = 0.000692331;	P[(int)('W' - 'A')][(int)('T' - 'A')] = -0.002995869;	P[(int)('W' - 'A')][(int)('D' - 'A')] = -0.012764388;	P[(int)('W' - 'A')][(int)('V' - 'A')] = -0.006705005;	P[(int)('W' - 'A')][(int)('A' - 'A')] = 0.006452907;	P[(int)('W' - 'A')][(int)('G' - 'A')] = 0.004774105;	P[(int)('W' - 'A')][(int)('R' - 'A')] = 0.001056551;	P[(int)('W' - 'A')][(int)('F' - 'A')] = 0.014735376;	P[(int)('W' - 'A')][(int)('P' - 'A')] = 0.000546254;	P[(int)('W' - 'A')][(int)('Q' - 'A')] = 0.005373195;	P[(int)('W' - 'A')][(int)('Y' - 'A')] = -0.011462277;	P[(int)('W' - 'A')][(int)('H' - 'A')] = 0.000555835;	P[(int)('W' - 'A')][(int)('M' - 'A')] = -0.032568334;	P[(int)('W' - 'A')][(int)('C' - 'A')] = 0.97378294;	P[(int)('W' - 'A')][(int)('W' - 'A')] = 0.2236068;
		
		P1[(int)('L' - 'A')][(int)('L' - 'A')] = -0.049681237;	P1[(int)('L' - 'A')][(int)('S' - 'A')] = 0.81716129;	P1[(int)('L' - 'A')][(int)('K' - 'A')] = 0.10517411;	P1[(int)('L' - 'A')][(int)('I' - 'A')] = 0.058081544;	P1[(int)('L' - 'A')][(int)('E' - 'A')] = -0.047533081;	P1[(int)('L' - 'A')][(int)('N' - 'A')] = -0.44696064;	P1[(int)('L' - 'A')][(int)('T' - 'A')] = -0.25165371;	P1[(int)('L' - 'A')][(int)('D' - 'A')] = 0.14269558;	P1[(int)('L' - 'A')][(int)('V' - 'A')] = -0.00671921;	P1[(int)('L' - 'A')][(int)('A' - 'A')] = -0.056616309;	P1[(int)('L' - 'A')][(int)('G' - 'A')] = -0.058954396;	P1[(int)('L' - 'A')][(int)('R' - 'A')] = -0.052229604;	P1[(int)('L' - 'A')][(int)('F' - 'A')] = -0.019895157;	P1[(int)('L' - 'A')][(int)('P' - 'A')] = -0.12562282;	P1[(int)('L' - 'A')][(int)('Q' - 'A')] = -0.000118842;	P1[(int)('L' - 'A')][(int)('Y' - 'A')] = -0.004703125;	P1[(int)('L' - 'A')][(int)('H' - 'A')] = 0.038982232;	P1[(int)('L' - 'A')][(int)('M' - 'A')] = 0.009115137;	P1[(int)('L' - 'A')][(int)('C' - 'A')] = -0.048576489;	P1[(int)('L' - 'A')][(int)('W' - 'A')] = -0.001945281;
		P1[(int)('S' - 'A')][(int)('L' - 'A')] = -0.002415121;	P1[(int)('S' - 'A')][(int)('S' - 'A')] = 0.061249572;	P1[(int)('S' - 'A')][(int)('K' - 'A')] = 0.00579841;	P1[(int)('S' - 'A')][(int)('I' - 'A')] = -0.59299062;	P1[(int)('S' - 'A')][(int)('E' - 'A')] = -0.014890516;	P1[(int)('S' - 'A')][(int)('N' - 'A')] = -0.074629599;	P1[(int)('S' - 'A')][(int)('T' - 'A')] = 0.28614434;	P1[(int)('S' - 'A')][(int)('D' - 'A')] = 0.028835636;	P1[(int)('S' - 'A')][(int)('V' - 'A')] = 0.64479343;	P1[(int)('S' - 'A')][(int)('A' - 'A')] = -0.37343765;	P1[(int)('S' - 'A')][(int)('G' - 'A')] = 0.010706691;	P1[(int)('S' - 'A')][(int)('R' - 'A')] = -0.006448166;	P1[(int)('S' - 'A')][(int)('F' - 'A')] = 0.001037422;	P1[(int)('S' - 'A')][(int)('P' - 'A')] = -0.002587705;	P1[(int)('S' - 'A')][(int)('Q' - 'A')] = 0.000522634;	P1[(int)('S' - 'A')][(int)('Y' - 'A')] = -0.00014131;	P1[(int)('S' - 'A')][(int)('H' - 'A')] = 0.007239687;	P1[(int)('S' - 'A')][(int)('M' - 'A')] = 0.025056105;	P1[(int)('S' - 'A')][(int)('C' - 'A')] = -0.003272976;	P1[(int)('S' - 'A')][(int)('W' - 'A')] = -0.000570264;
		P1[(int)('K' - 'A')][(int)('L' - 'A')] = -0.017779188;	P1[(int)('K' - 'A')][(int)('S' - 'A')] = 0.15587172;	P1[(int)('K' - 'A')][(int)('K' - 'A')] = -0.15827075;	P1[(int)('K' - 'A')][(int)('I' - 'A')] = -0.25514888;	P1[(int)('K' - 'A')][(int)('E' - 'A')] = 0.15818656;	P1[(int)('K' - 'A')][(int)('N' - 'A')] = 0.40818135;	P1[(int)('K' - 'A')][(int)('T' - 'A')] = -0.61615283;	P1[(int)('K' - 'A')][(int)('D' - 'A')] = -0.29242293;	P1[(int)('K' - 'A')][(int)('V' - 'A')] = 0.29093514;	P1[(int)('K' - 'A')][(int)('A' - 'A')] = 0.34804428;	P1[(int)('K' - 'A')][(int)('G' - 'A')] = -0.064696527;	P1[(int)('K' - 'A')][(int)('R' - 'A')] = 0.097640225;	P1[(int)('K' - 'A')][(int)('F' - 'A')] = -0.002052849;	P1[(int)('K' - 'A')][(int)('P' - 'A')] = -0.030032023;	P1[(int)('K' - 'A')][(int)('Q' - 'A')] = -0.010300904;	P1[(int)('K' - 'A')][(int)('Y' - 'A')] = -0.000049959719;	P1[(int)('K' - 'A')][(int)('H' - 'A')] = -0.060140373;	P1[(int)('K' - 'A')][(int)('M' - 'A')] = 0.06398273;	P1[(int)('K' - 'A')][(int)('C' - 'A')] = -0.015034294;	P1[(int)('K' - 'A')][(int)('W' - 'A')] = -0.000760496;
		P1[(int)('I' - 'A')][(int)('L' - 'A')] = -0.026492808;	P1[(int)('I' - 'A')][(int)('S' - 'A')] = 0.31333291;	P1[(int)('I' - 'A')][(int)('K' - 'A')] = -0.41563643;	P1[(int)('I' - 'A')][(int)('I' - 'A')] = 0.19477615;	P1[(int)('I' - 'A')][(int)('E' - 'A')] = 0.38193051;	P1[(int)('I' - 'A')][(int)('N' - 'A')] = 0.22989943;	P1[(int)('I' - 'A')][(int)('T' - 'A')] = 0.27315449;	P1[(int)('I' - 'A')][(int)('D' - 'A')] = -0.38034106;	P1[(int)('I' - 'A')][(int)('V' - 'A')] = -0.15579176;	P1[(int)('I' - 'A')][(int)('A' - 'A')] = -0.42188597;	P1[(int)('I' - 'A')][(int)('G' - 'A')] = -0.021783595;	P1[(int)('I' - 'A')][(int)('R' - 'A')] = 0.24537214;	P1[(int)('I' - 'A')][(int)('F' - 'A')] = -0.010189739;	P1[(int)('I' - 'A')][(int)('P' - 'A')] = -0.051698466;	P1[(int)('I' - 'A')][(int)('Q' - 'A')] = -0.03107875;	P1[(int)('I' - 'A')][(int)('Y' - 'A')] = 0.002107193;	P1[(int)('I' - 'A')][(int)('H' - 'A')] = -0.062028214;	P1[(int)('I' - 'A')][(int)('M' - 'A')] = -0.029439419;	P1[(int)('I' - 'A')][(int)('C' - 'A')] = -0.032018034;	P1[(int)('I' - 'A')][(int)('W' - 'A')] = -0.002188577;
		P1[(int)('E' - 'A')][(int)('L' - 'A')] = 0.003190431;	P1[(int)('E' - 'A')][(int)('S' - 'A')] = 0.066562022;	P1[(int)('E' - 'A')][(int)('K' - 'A')] = 0.59004463;	P1[(int)('E' - 'A')][(int)('I' - 'A')] = 0.063859591;	P1[(int)('E' - 'A')][(int)('E' - 'A')] = 0.19960266;	P1[(int)('E' - 'A')][(int)('N' - 'A')] = 0.1947956;	P1[(int)('E' - 'A')][(int)('T' - 'A')] = 0.007468631;	P1[(int)('E' - 'A')][(int)('D' - 'A')] = -0.42496808;	P1[(int)('E' - 'A')][(int)('V' - 'A')] = -0.015026826;	P1[(int)('E' - 'A')][(int)('A' - 'A')] = -0.16969608;	P1[(int)('E' - 'A')][(int)('G' - 'A')] = 0.05523449;	P1[(int)('E' - 'A')][(int)('R' - 'A')] = -0.57456348;	P1[(int)('E' - 'A')][(int)('F' - 'A')] = -0.002269018;	P1[(int)('E' - 'A')][(int)('P' - 'A')] = 0.01066163;	P1[(int)('E' - 'A')][(int)('Q' - 'A')] = -0.089795219;	P1[(int)('E' - 'A')][(int)('Y' - 'A')] = -0.015299426;	P1[(int)('E' - 'A')][(int)('H' - 'A')] = 0.11571894;	P1[(int)('E' - 'A')][(int)('M' - 'A')] = -0.028172138;	P1[(int)('E' - 'A')][(int)('C' - 'A')] = 0.008686522;	P1[(int)('E' - 'A')][(int)('W' - 'A')] = 0.003965127;
		P1[(int)('N' - 'A')][(int)('L' - 'A')] = -0.065180382;	P1[(int)('N' - 'A')][(int)('S' - 'A')] = 0.15059126;	P1[(int)('N' - 'A')][(int)('K' - 'A')] = 0.045540384;	P1[(int)('N' - 'A')][(int)('I' - 'A')] = 0.18749096;	P1[(int)('N' - 'A')][(int)('E' - 'A')] = -0.56531243;	P1[(int)('N' - 'A')][(int)('N' - 'A')] = 0.53009713;	P1[(int)('N' - 'A')][(int)('T' - 'A')] = -0.11026103;	P1[(int)('N' - 'A')][(int)('D' - 'A')] = 0.19846483;	P1[(int)('N' - 'A')][(int)('V' - 'A')] = 0.044523791;	P1[(int)('N' - 'A')][(int)('A' - 'A')] = -0.35211762;	P1[(int)('N' - 'A')][(int)('G' - 'A')] = 0.027676821;	P1[(int)('N' - 'A')][(int)('R' - 'A')] = -0.072144571;	P1[(int)('N' - 'A')][(int)('F' - 'A')] = -0.00604177;	P1[(int)('N' - 'A')][(int)('P' - 'A')] = -0.001360316;	P1[(int)('N' - 'A')][(int)('Q' - 'A')] = 0.27649482;	P1[(int)('N' - 'A')][(int)('Y' - 'A')] = 0.028701239;	P1[(int)('N' - 'A')][(int)('H' - 'A')] = -0.27078197;	P1[(int)('N' - 'A')][(int)('M' - 'A')] = -0.027853075;	P1[(int)('N' - 'A')][(int)('C' - 'A')] = -0.019220409;	P1[(int)('N' - 'A')][(int)('W' - 'A')] = 0.000692331;
		P1[(int)('T' - 'A')][(int)('L' - 'A')] = 0.22015554;	P1[(int)('T' - 'A')][(int)('S' - 'A')] = 0.23673576;	P1[(int)('T' - 'A')][(int)('K' - 'A')] = 0.002593839;	P1[(int)('T' - 'A')][(int)('I' - 'A')] = -0.41081291;	P1[(int)('T' - 'A')][(int)('E' - 'A')] = -0.12978641;	P1[(int)('T' - 'A')][(int)('N' - 'A')] = 0.15888675;	P1[(int)('T' - 'A')][(int)('T' - 'A')] = 0.43010446;	P1[(int)('T' - 'A')][(int)('D' - 'A')] = -0.09024041;	P1[(int)('T' - 'A')][(int)('V' - 'A')] = -0.34448917;	P1[(int)('T' - 'A')][(int)('A' - 'A')] = 0.39387541;	P1[(int)('T' - 'A')][(int)('G' - 'A')] = -0.15996143;	P1[(int)('T' - 'A')][(int)('R' - 'A')] = -0.13903561;	P1[(int)('T' - 'A')][(int)('F' - 'A')] = -0.040261502;	P1[(int)('T' - 'A')][(int)('P' - 'A')] = -0.27446871;	P1[(int)('T' - 'A')][(int)('Q' - 'A')] = 0.22743903;	P1[(int)('T' - 'A')][(int)('Y' - 'A')] = 0.041402345;	P1[(int)('T' - 'A')][(int)('H' - 'A')] = -0.18122503;	P1[(int)('T' - 'A')][(int)('M' - 'A')] = 0.092631083;	P1[(int)('T' - 'A')][(int)('C' - 'A')] = -0.030547168;	P1[(int)('T' - 'A')][(int)('W' - 'A')] = -0.002995869;
		P1[(int)('D' - 'A')][(int)('L' - 'A')] = 0.73416107;	P1[(int)('D' - 'A')][(int)('S' - 'A')] = 0.040072747;	P1[(int)('D' - 'A')][(int)('K' - 'A')] = 0.086828695;	P1[(int)('D' - 'A')][(int)('I' - 'A')] = -0.00201392;	P1[(int)('D' - 'A')][(int)('E' - 'A')] = -0.044030883;	P1[(int)('D' - 'A')][(int)('N' - 'A')] = 0.15108131;	P1[(int)('D' - 'A')][(int)('T' - 'A')] = -0.075518941;	P1[(int)('D' - 'A')][(int)('D' - 'A')] = 0.14867582;	P1[(int)('D' - 'A')][(int)('V' - 'A')] = -0.001989732;	P1[(int)('D' - 'A')][(int)('A' - 'A')] = -0.1017408;	P1[(int)('D' - 'A')][(int)('G' - 'A')] = -0.13748663;	P1[(int)('D' - 'A')][(int)('R' - 'A')] = 0.14098981;	P1[(int)('D' - 'A')][(int)('F' - 'A')] = -0.20174552;	P1[(int)('D' - 'A')][(int)('P' - 'A')] = -0.058605507;	P1[(int)('D' - 'A')][(int)('Q' - 'A')] = -0.38928252;	P1[(int)('D' - 'A')][(int)('Y' - 'A')] = 0.004054481;	P1[(int)('D' - 'A')][(int)('H' - 'A')] = 0.11423333;	P1[(int)('D' - 'A')][(int)('M' - 'A')] = -0.37531455;	P1[(int)('D' - 'A')][(int)('C' - 'A')] = -0.019603867;	P1[(int)('D' - 'A')][(int)('W' - 'A')] = -0.012764388;
		P1[(int)('V' - 'A')][(int)('L' - 'A')] = 0.29055798;	P1[(int)('V' - 'A')][(int)('S' - 'A')] = -0.10097594;	P1[(int)('V' - 'A')][(int)('K' - 'A')] = 0.026265838;	P1[(int)('V' - 'A')][(int)('I' - 'A')] = 0.11120814;	P1[(int)('V' - 'A')][(int)('E' - 'A')] = 0.15943821;	P1[(int)('V' - 'A')][(int)('N' - 'A')] = -0.30671582;	P1[(int)('V' - 'A')][(int)('T' - 'A')] = -0.09575316;	P1[(int)('V' - 'A')][(int)('D' - 'A')] = -0.19645181;	P1[(int)('V' - 'A')][(int)('V' - 'A')] = 0.150836;	P1[(int)('V' - 'A')][(int)('A' - 'A')] = 0.010508699;	P1[(int)('V' - 'A')][(int)('G' - 'A')] = 0.26397134;	P1[(int)('V' - 'A')][(int)('R' - 'A')] = -0.024022526;	P1[(int)('V' - 'A')][(int)('F' - 'A')] = -0.14090331;	P1[(int)('V' - 'A')][(int)('P' - 'A')] = -0.00208106;	P1[(int)('V' - 'A')][(int)('Q' - 'A')] = 0.44373145;	P1[(int)('V' - 'A')][(int)('Y' - 'A')] = 0.22430225;	P1[(int)('V' - 'A')][(int)('H' - 'A')] = -0.52948374;	P1[(int)('V' - 'A')][(int)('M' - 'A')] = -0.29037999;	P1[(int)('V' - 'A')][(int)('C' - 'A')] = 0.012652466;	P1[(int)('V' - 'A')][(int)('W' - 'A')] = -0.006705005;
		P1[(int)('A' - 'A')][(int)('L' - 'A')] = -0.18070275;	P1[(int)('A' - 'A')][(int)('S' - 'A')] = -0.031716243;	P1[(int)('A' - 'A')][(int)('K' - 'A')] = 0.15779057;	P1[(int)('A' - 'A')][(int)('I' - 'A')] = 0.12305416;	P1[(int)('A' - 'A')][(int)('E' - 'A')] = 0.35145129;	P1[(int)('A' - 'A')][(int)('N' - 'A')] = 0.070081118;	P1[(int)('A' - 'A')][(int)('T' - 'A')] = 0.070162308;	P1[(int)('A' - 'A')][(int)('D' - 'A')] = 0.2957769;	P1[(int)('A' - 'A')][(int)('V' - 'A')] = 0.1265477;	P1[(int)('A' - 'A')][(int)('A' - 'A')] = 0.03547992;	P1[(int)('A' - 'A')][(int)('G' - 'A')] = -0.7291697;	P1[(int)('A' - 'A')][(int)('R' - 'A')] = -0.049717429;	P1[(int)('A' - 'A')][(int)('F' - 'A')] = 0.011250757;	P1[(int)('A' - 'A')][(int)('P' - 'A')] = 0.071722469;	P1[(int)('A' - 'A')][(int)('Q' - 'A')] = -0.01490388;	P1[(int)('A' - 'A')][(int)('Y' - 'A')] = 0.10998942;	P1[(int)('A' - 'A')][(int)('H' - 'A')] = -0.35321224;	P1[(int)('A' - 'A')][(int)('M' - 'A')] = -0.10286728;	P1[(int)('A' - 'A')][(int)('C' - 'A')] = 0.032529991;	P1[(int)('A' - 'A')][(int)('W' - 'A')] = 0.006452907;
		P1[(int)('G' - 'A')][(int)('L' - 'A')] = -0.12485798;	P1[(int)('G' - 'A')][(int)('S' - 'A')] = 0.032901684;	P1[(int)('G' - 'A')][(int)('K' - 'A')] = -0.20092364;	P1[(int)('G' - 'A')][(int)('I' - 'A')] = 0.28544197;	P1[(int)('G' - 'A')][(int)('E' - 'A')] = -0.12259817;	P1[(int)('G' - 'A')][(int)('N' - 'A')] = 0.026053068;	P1[(int)('G' - 'A')][(int)('T' - 'A')] = 0.19633815;	P1[(int)('G' - 'A')][(int)('D' - 'A')] = -0.10008594;	P1[(int)('G' - 'A')][(int)('V' - 'A')] = 0.34671352;	P1[(int)('G' - 'A')][(int)('A' - 'A')] = 0.26407098;	P1[(int)('G' - 'A')][(int)('G' - 'A')] = -0.11893473;	P1[(int)('G' - 'A')][(int)('R' - 'A')] = -0.16132325;	P1[(int)('G' - 'A')][(int)('F' - 'A')] = 0.044979776;	P1[(int)('G' - 'A')][(int)('P' - 'A')] = -0.26755618;	P1[(int)('G' - 'A')][(int)('Q' - 'A')] = 0.21755774;	P1[(int)('G' - 'A')][(int)('Y' - 'A')] = -0.16526757;	P1[(int)('G' - 'A')][(int)('H' - 'A')] = 0.36961101;	P1[(int)('G' - 'A')][(int)('M' - 'A')] = -0.53120924;	P1[(int)('G' - 'A')][(int)('C' - 'A')] = 0.00431469;	P1[(int)('G' - 'A')][(int)('W' - 'A')] = 0.004774105;
		P1[(int)('R' - 'A')][(int)('L' - 'A')] = 0.28014971;	P1[(int)('R' - 'A')][(int)('S' - 'A')] = -0.070946219;	P1[(int)('R' - 'A')][(int)('K' - 'A')] = -0.39370815;	P1[(int)('R' - 'A')][(int)('I' - 'A')] = -0.071589204;	P1[(int)('R' - 'A')][(int)('E' - 'A')] = 0.26434271;	P1[(int)('R' - 'A')][(int)('N' - 'A')] = 0.013175221;	P1[(int)('R' - 'A')][(int)('T' - 'A')] = -0.2000809;	P1[(int)('R' - 'A')][(int)('D' - 'A')] = 0.28964442;	P1[(int)('R' - 'A')][(int)('V' - 'A')] = -0.098619711;	P1[(int)('R' - 'A')][(int)('A' - 'A')] = -0.19780946;	P1[(int)('R' - 'A')][(int)('G' - 'A')] = -0.092767701;	P1[(int)('R' - 'A')][(int)('R' - 'A')] = -0.49792519;	P1[(int)('R' - 'A')][(int)('F' - 'A')] = -0.046092584;	P1[(int)('R' - 'A')][(int)('P' - 'A')] = 0.071077241;	P1[(int)('R' - 'A')][(int)('Q' - 'A')] = 0.36884631;	P1[(int)('R' - 'A')][(int)('Y' - 'A')] = -0.11546305;	P1[(int)('R' - 'A')][(int)('H' - 'A')] = 0.25098882;	P1[(int)('R' - 'A')][(int)('M' - 'A')] = 0.18304029;	P1[(int)('R' - 'A')][(int)('C' - 'A')] = 0.06268089;	P1[(int)('R' - 'A')][(int)('W' - 'A')] = 0.001056551;
		P1[(int)('F' - 'A')][(int)('L' - 'A')] = -0.1813374;	P1[(int)('F' - 'A')][(int)('S' - 'A')] = 0.065944418;	P1[(int)('F' - 'A')][(int)('K' - 'A')] = -0.24896154;	P1[(int)('F' - 'A')][(int)('I' - 'A')] = 0.076413924;	P1[(int)('F' - 'A')][(int)('E' - 'A')] = 0.070585802;	P1[(int)('F' - 'A')][(int)('N' - 'A')] = 0.11931146;	P1[(int)('F' - 'A')][(int)('T' - 'A')] = 0.11319097;	P1[(int)('F' - 'A')][(int)('D' - 'A')] = 0.23053356;	P1[(int)('F' - 'A')][(int)('V' - 'A')] = 0.10119853;	P1[(int)('F' - 'A')][(int)('A' - 'A')] = 0.15314603;	P1[(int)('F' - 'A')][(int)('G' - 'A')] = 0.29056045;	P1[(int)('F' - 'A')][(int)('R' - 'A')] = -0.33609857;	P1[(int)('F' - 'A')][(int)('F' - 'A')] = -0.37647547;	P1[(int)('F' - 'A')][(int)('P' - 'A')] = -0.081213172;	P1[(int)('F' - 'A')][(int)('Q' - 'A')] = -0.35675943;	P1[(int)('F' - 'A')][(int)('Y' - 'A')] = 0.52939247;	P1[(int)('F' - 'A')][(int)('H' - 'A')] = -0.053141185;	P1[(int)('F' - 'A')][(int)('M' - 'A')] = 0.009323667;	P1[(int)('F' - 'A')][(int)('C' - 'A')] = -0.14034989;	P1[(int)('F' - 'A')][(int)('W' - 'A')] = 0.014735376;
		P1[(int)('P' - 'A')][(int)('L' - 'A')] = -0.012401368;	P1[(int)('P' - 'A')][(int)('S' - 'A')] = -0.036255272;	P1[(int)('P' - 'A')][(int)('K' - 'A')] = 0.028638242;	P1[(int)('P' - 'A')][(int)('I' - 'A')] = -0.017739701;	P1[(int)('P' - 'A')][(int)('E' - 'A')] = 0.26961129;	P1[(int)('P' - 'A')][(int)('N' - 'A')] = 0.1047864;	P1[(int)('P' - 'A')][(int)('T' - 'A')] = -0.038537922;	P1[(int)('P' - 'A')][(int)('D' - 'A')] = 0.31858099;	P1[(int)('P' - 'A')][(int)('V' - 'A')] = -0.006939004;	P1[(int)('P' - 'A')][(int)('A' - 'A')] = -0.021309736;	P1[(int)('P' - 'A')][(int)('G' - 'A')] = 0.33461851;	P1[(int)('P' - 'A')][(int)('R' - 'A')] = -0.070869891;	P1[(int)('P' - 'A')][(int)('F' - 'A')] = 0.45009843;	P1[(int)('P' - 'A')][(int)('P' - 'A')] = -0.54047866;	P1[(int)('P' - 'A')][(int)('Q' - 'A')] = -0.17926459;	P1[(int)('P' - 'A')][(int)('Y' - 'A')] = -0.28923863;	P1[(int)('P' - 'A')][(int)('H' - 'A')] = -0.27033524;	P1[(int)('P' - 'A')][(int)('M' - 'A')] = -0.080292469;	P1[(int)('P' - 'A')][(int)('C' - 'A')] = 0.056782361;	P1[(int)('P' - 'A')][(int)('W' - 'A')] = 0.000546254;
		P1[(int)('Q' - 'A')][(int)('L' - 'A')] = -0.046902305;	P1[(int)('Q' - 'A')][(int)('S' - 'A')] = 0.19477143;	P1[(int)('Q' - 'A')][(int)('K' - 'A')] = -0.15548169;	P1[(int)('Q' - 'A')][(int)('I' - 'A')] = -0.17796799;	P1[(int)('Q' - 'A')][(int)('E' - 'A')] = -0.000086778831;	P1[(int)('Q' - 'A')][(int)('N' - 'A')] = 0.048731598;	P1[(int)('Q' - 'A')][(int)('T' - 'A')] = 0.080977783;	P1[(int)('Q' - 'A')][(int)('D' - 'A')] = 0.070353596;	P1[(int)('Q' - 'A')][(int)('V' - 'A')] = -0.11954331;	P1[(int)('Q' - 'A')][(int)('A' - 'A')] = 0.1374577;	P1[(int)('Q' - 'A')][(int)('G' - 'A')] = 0.16088686;	P1[(int)('Q' - 'A')][(int)('R' - 'A')] = -0.19068849;	P1[(int)('Q' - 'A')][(int)('F' - 'A')] = 0.29157229;	P1[(int)('Q' - 'A')][(int)('P' - 'A')] = 0.67396246;	P1[(int)('Q' - 'A')][(int)('Q' - 'A')] = -0.13555852;	P1[(int)('Q' - 'A')][(int)('Y' - 'A')] = -0.18324817;	P1[(int)('Q' - 'A')][(int)('H' - 'A')] = -0.16669016;	P1[(int)('Q' - 'A')][(int)('M' - 'A')] = -0.41227551;	P1[(int)('Q' - 'A')][(int)('C' - 'A')] = -0.075644002;	P1[(int)('Q' - 'A')][(int)('W' - 'A')] = 0.005373195;
		P1[(int)('Y' - 'A')][(int)('L' - 'A')] = -0.035150013;	P1[(int)('Y' - 'A')][(int)('S' - 'A')] = 0.079643091;	P1[(int)('Y' - 'A')][(int)('K' - 'A')] = -0.07368193;	P1[(int)('Y' - 'A')][(int)('I' - 'A')] = 0.067092985;	P1[(int)('Y' - 'A')][(int)('E' - 'A')] = -0.05864428;	P1[(int)('Y' - 'A')][(int)('N' - 'A')] = -0.011323853;	P1[(int)('Y' - 'A')][(int)('T' - 'A')] = 0.074447331;	P1[(int)('Y' - 'A')][(int)('D' - 'A')] = -0.030264684;	P1[(int)('Y' - 'A')][(int)('V' - 'A')] = 0.070286963;	P1[(int)('Y' - 'A')][(int)('A' - 'A')] = 0.08232548;	P1[(int)('Y' - 'A')][(int)('G' - 'A')] = 0.053018954;	P1[(int)('Y' - 'A')][(int)('R' - 'A')] = -0.074658126;	P1[(int)('Y' - 'A')][(int)('F' - 'A')] = -0.3754672;	P1[(int)('Y' - 'A')][(int)('P' - 'A')] = 0.060779576;	P1[(int)('Y' - 'A')][(int)('Q' - 'A')] = -0.15474384;	P1[(int)('Y' - 'A')][(int)('Y' - 'A')] = -0.34477172;	P1[(int)('Y' - 'A')][(int)('H' - 'A')] = -0.18630698;	P1[(int)('Y' - 'A')][(int)('M' - 'A')] = 0.079080378;	P1[(int)('Y' - 'A')][(int)('C' - 'A')] = 0.78980015;	P1[(int)('Y' - 'A')][(int)('W' - 'A')] = -0.011462277;
		P1[(int)('H' - 'A')][(int)('L' - 'A')] = -0.32385684;	P1[(int)('H' - 'A')][(int)('S' - 'A')] = -0.001441757;	P1[(int)('H' - 'A')][(int)('K' - 'A')] = 0.20682824;	P1[(int)('H' - 'A')][(int)('I' - 'A')] = -0.34091978;	P1[(int)('H' - 'A')][(int)('E' - 'A')] = 0.21638688;	P1[(int)('H' - 'A')][(int)('N' - 'A')] = 0.13459322;	P1[(int)('H' - 'A')][(int)('T' - 'A')] = -0.11512003;	P1[(int)('H' - 'A')][(int)('D' - 'A')] = 0.20105761;	P1[(int)('H' - 'A')][(int)('V' - 'A')] = -0.30566775;	P1[(int)('H' - 'A')][(int)('A' - 'A')] = -0.11543601;	P1[(int)('H' - 'A')][(int)('G' - 'A')] = 0.13845924;	P1[(int)('H' - 'A')][(int)('R' - 'A')] = 0.22846803;	P1[(int)('H' - 'A')][(int)('F' - 'A')] = -0.26779468;	P1[(int)('H' - 'A')][(int)('P' - 'A')] = -0.057459336;	P1[(int)('H' - 'A')][(int)('Q' - 'A')] = 0.25156734;	P1[(int)('H' - 'A')][(int)('Y' - 'A')] = 0.13100243;	P1[(int)('H' - 'A')][(int)('H' - 'A')] = 0.24147237;	P1[(int)('H' - 'A')][(int)('M' - 'A')] = -0.42820641;	P1[(int)('H' - 'A')][(int)('C' - 'A')] = 0.2055114;	P1[(int)('H' - 'A')][(int)('W' - 'A')] = 0.000555835;
		P1[(int)('M' - 'A')][(int)('L' - 'A')] = 0.03378741;	P1[(int)('M' - 'A')][(int)('S' - 'A')] = -0.037550607;	P1[(int)('M' - 'A')][(int)('K' - 'A')] = -0.15559485;	P1[(int)('M' - 'A')][(int)('I' - 'A')] = -0.089891867;	P1[(int)('M' - 'A')][(int)('E' - 'A')] = -0.18148197;	P1[(int)('M' - 'A')][(int)('N' - 'A')] = -0.095339243;	P1[(int)('M' - 'A')][(int)('T' - 'A')] = -0.10933631;	P1[(int)('M' - 'A')][(int)('D' - 'A')] = -0.15388422;	P1[(int)('M' - 'A')][(int)('V' - 'A')] = -0.097616607;	P1[(int)('M' - 'A')][(int)('A' - 'A')] = -0.11560482;	P1[(int)('M' - 'A')][(int)('G' - 'A')] = -0.14782752;	P1[(int)('M' - 'A')][(int)('R' - 'A')] = -0.1090617;	P1[(int)('M' - 'A')][(int)('F' - 'A')] = 0.48930639;	P1[(int)('M' - 'A')][(int)('P' - 'A')] = -0.082476431;	P1[(int)('M' - 'A')][(int)('Q' - 'A')] = -0.11299245;	P1[(int)('M' - 'A')][(int)('Y' - 'A')] = 0.55575385;	P1[(int)('M' - 'A')][(int)('H' - 'A')] = 0.064615376;	P1[(int)('M' - 'A')][(int)('M' - 'A')] = -0.11641769;	P1[(int)('M' - 'A')][(int)('C' - 'A')] = 0.49418157;	P1[(int)('M' - 'A')][(int)('W' - 'A')] = -0.032568334;
		P1[(int)('C' - 'A')][(int)('L' - 'A')] = -0.033768978;	P1[(int)('C' - 'A')][(int)('S' - 'A')] = -0.050915707;	P1[(int)('C' - 'A')][(int)('K' - 'A')] = -0.054511853;	P1[(int)('C' - 'A')][(int)('I' - 'A')] = -0.056580404;	P1[(int)('C' - 'A')][(int)('E' - 'A')] = -0.060874437;	P1[(int)('C' - 'A')][(int)('N' - 'A')] = -0.0583983;	P1[(int)('C' - 'A')][(int)('T' - 'A')] = -0.057889781;	P1[(int)('C' - 'A')][(int)('D' - 'A')] = -0.06167617;	P1[(int)('C' - 'A')][(int)('V' - 'A')] = -0.056684905;	P1[(int)('C' - 'A')][(int)('A' - 'A')] = -0.058673043;	P1[(int)('C' - 'A')][(int)('G' - 'A')] = -0.056634067;	P1[(int)('C' - 'A')][(int)('R' - 'A')] = -0.043999086;	P1[(int)('C' - 'A')][(int)('F' - 'A')] = -0.039446127;	P1[(int)('C' - 'A')][(int)('P' - 'A')] = -0.05684524;	P1[(int)('C' - 'A')][(int)('Q' - 'A')] = -0.053467688;	P1[(int)('C' - 'A')][(int)('Y' - 'A')] = -0.041760599;	P1[(int)('C' - 'A')][(int)('H' - 'A')] = -0.052473005;	P1[(int)('C' - 'A')][(int)('M' - 'A')] = -0.055281138;	P1[(int)('C' - 'A')][(int)('C' - 'A')] = -0.023902416;	P1[(int)('C' - 'A')][(int)('W' - 'A')] = 0.97378294;
		P1[(int)('W' - 'A')][(int)('L' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('S' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('K' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('I' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('E' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('N' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('T' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('D' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('V' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('A' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('G' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('R' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('F' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('P' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('Q' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('Y' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('H' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('M' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('C' - 'A')] = 0.2236068;	P1[(int)('W' - 'A')][(int)('W' - 'A')] = 0.2236068;
	
		sum_letters[(int)('L' - 'A')]  = 0.0963342053386846;
		sum_letters[(int)('S' - 'A')]  = 0.0882158107506429;
		sum_letters[(int)('K' - 'A')]  = 0.0732072974102184;
		sum_letters[(int)('I' - 'A')]  = 0.0637063066162858;
		sum_letters[(int)('E' - 'A')]  = 0.0660862696594149;
		sum_letters[(int)('N' - 'A')]  = 0.0602300973875848;
		sum_letters[(int)('T' - 'A')]  = 0.0577678301469987;
		sum_letters[(int)('D' - 'A')]  = 0.0608816722841187;
		sum_letters[(int)('V' - 'A')]  = 0.0562852114789207;
		sum_letters[(int)('A' - 'A')]  = 0.0558982674394878;
		sum_letters[(int)('G' - 'A')]  = 0.0505473514909348;
		sum_letters[(int)('R' - 'A')]  = 0.0444825609408284;
		sum_letters[(int)('F' - 'A')]  = 0.0441144782799794;
		sum_letters[(int)('P' - 'A')]  = 0.0442087851728988;
		sum_letters[(int)('Q' - 'A')]  = 0.0403324860954489;
		sum_letters[(int)('Y' - 'A')]  = 0.0326479032148362;
		sum_letters[(int)('H' - 'A')]  = 0.0216785826759952;
		sum_letters[(int)('M' - 'A')]  = 0.0207315128483139;
		sum_letters[(int)('C' - 'A')]  = 0.0122581814087392;
		sum_letters[(int)('W' - 'A')]  = 0.0103851893596676;
	}
	else{
		lambda[(int)('L' - 'A')]  = -0.189812092944174;
		lambda[(int)('S' - 'A')]  = -3.00000044953125E-07;
		lambda[(int)('K' - 'A')]  = -0.161670618826804;
		lambda[(int)('I' - 'A')]  = -0.157214376490117;
		lambda[(int)('E' - 'A')]  = -0.152909355589079;
		lambda[(int)('N' - 'A')]  = -0.0253915748075592;
		lambda[(int)('T' - 'A')]  = -0.137019870293421;
		lambda[(int)('D' - 'A')]  = -0.0340912075902619;
		lambda[(int)('V' - 'A')]  = -0.0409734018150945;
		lambda[(int)('A' - 'A')]  = -0.130538401807701;
		lambda[(int)('G' - 'A')]  = -0.119324019862902;
		lambda[(int)('R' - 'A')]  = -0.0495939706353095;
		lambda[(int)('F' - 'A')]  = -0.0547153851720098;
		lambda[(int)('P' - 'A')]  = -0.0587445016006264;
		lambda[(int)('Q' - 'A')]  = -0.107462334042498;
		lambda[(int)('Y' - 'A')]  = -0.074066423786098;
		lambda[(int)('H' - 'A')]  = -0.0980855756547957;
		lambda[(int)('M' - 'A')]  = -0.0962850778630469;
		lambda[(int)('C' - 'A')]  = -0.0858769915572378;
		
		P[(int)('L' - 'A')][(int)('L' - 'A')] =0.091054625;	P[(int)('L' - 'A')][(int)('S' - 'A')] =-0.22360951;	P[(int)('L' - 'A')][(int)('K' - 'A')] =0.70545772;	P[(int)('L' - 'A')][(int)('I' - 'A')] =0.23883865;	P[(int)('L' - 'A')][(int)('E' - 'A')] =0.13368332;	P[(int)('L' - 'A')][(int)('N' - 'A')] =0.059742572;	P[(int)('L' - 'A')][(int)('T' - 'A')] =-0.28980527;	P[(int)('L' - 'A')][(int)('D' - 'A')] =0.18080128;	P[(int)('L' - 'A')][(int)('V' - 'A')] =0.26339188;	P[(int)('L' - 'A')][(int)('A' - 'A')] =-0.025532082;	P[(int)('L' - 'A')][(int)('G' - 'A')] =0.072476512;	P[(int)('L' - 'A')][(int)('R' - 'A')] =-0.14277194;	P[(int)('L' - 'A')][(int)('F' - 'A')] =0.037873135;	P[(int)('L' - 'A')][(int)('P' - 'A')] =-0.031140868;	P[(int)('L' - 'A')][(int)('Q' - 'A')] =0.039731329;	P[(int)('L' - 'A')][(int)('Y' - 'A')] =0.059903092;	P[(int)('L' - 'A')][(int)('H' - 'A')] =0.25149469;	P[(int)('L' - 'A')][(int)('M' - 'A')] =-0.082064798;	P[(int)('L' - 'A')][(int)('C' - 'A')] =-0.27492555;	P[(int)('L' - 'A')][(int)('W' - 'A')] =-0.010241735;
		P[(int)('S' - 'A')][(int)('L' - 'A')] =-0.50203557;	P[(int)('S' - 'A')][(int)('S' - 'A')] =-0.22360826;	P[(int)('S' - 'A')][(int)('K' - 'A')] =0.28273666;	P[(int)('S' - 'A')][(int)('I' - 'A')] =-0.54840356;	P[(int)('S' - 'A')][(int)('E' - 'A')] =-0.28889045;	P[(int)('S' - 'A')][(int)('N' - 'A')] =0.022198317;	P[(int)('S' - 'A')][(int)('T' - 'A')] =-0.029974738;	P[(int)('S' - 'A')][(int)('D' - 'A')] =-0.09897819;	P[(int)('S' - 'A')][(int)('V' - 'A')] =-0.046088582;	P[(int)('S' - 'A')][(int)('A' - 'A')] =-0.11480396;	P[(int)('S' - 'A')][(int)('G' - 'A')] =0.10005415;	P[(int)('S' - 'A')][(int)('R' - 'A')] =-0.026084326;	P[(int)('S' - 'A')][(int)('F' - 'A')] =-0.013851001;	P[(int)('S' - 'A')][(int)('P' - 'A')] =0.08621139;	P[(int)('S' - 'A')][(int)('Q' - 'A')] =-0.042957423;	P[(int)('S' - 'A')][(int)('Y' - 'A')] =-0.080001047;	P[(int)('S' - 'A')][(int)('H' - 'A')] =0.15883087;	P[(int)('S' - 'A')][(int)('M' - 'A')] =-0.25619551;	P[(int)('S' - 'A')][(int)('C' - 'A')] =0.29111561;	P[(int)('S' - 'A')][(int)('W' - 'A')] =-0.071479025;
		P[(int)('K' - 'A')][(int)('L' - 'A')] =-0.017283163;	P[(int)('K' - 'A')][(int)('S' - 'A')] =-0.22360515;	P[(int)('K' - 'A')][(int)('K' - 'A')] =0.056612688;	P[(int)('K' - 'A')][(int)('I' - 'A')] =-0.1572796;	P[(int)('K' - 'A')][(int)('E' - 'A')] =0.22563233;	P[(int)('K' - 'A')][(int)('N' - 'A')] =0.058778297;	P[(int)('K' - 'A')][(int)('T' - 'A')] =0.0037734757;	P[(int)('K' - 'A')][(int)('D' - 'A')] =-0.14655599;	P[(int)('K' - 'A')][(int)('V' - 'A')] =-0.066022924;	P[(int)('K' - 'A')][(int)('A' - 'A')] =0.73775567;	P[(int)('K' - 'A')][(int)('G' - 'A')] =-0.35139616;	P[(int)('K' - 'A')][(int)('R' - 'A')] =0.0011327131;	P[(int)('K' - 'A')][(int)('F' - 'A')] =0.089835679;	P[(int)('K' - 'A')][(int)('P' - 'A')] =-0.14510671;	P[(int)('K' - 'A')][(int)('Q' - 'A')] =0.025292546;	P[(int)('K' - 'A')][(int)('Y' - 'A')] =0.1627435;	P[(int)('K' - 'A')][(int)('H' - 'A')] =0.012136342;	P[(int)('K' - 'A')][(int)('M' - 'A')] =0.040934955;	P[(int)('K' - 'A')][(int)('C' - 'A')] =0.06142347;	P[(int)('K' - 'A')][(int)('W' - 'A')] =-0.33455887;
		P[(int)('I' - 'A')][(int)('L' - 'A')] =0.087740566;	P[(int)('I' - 'A')][(int)('S' - 'A')] =-0.22360959;	P[(int)('I' - 'A')][(int)('K' - 'A')] =0.16731999;	P[(int)('I' - 'A')][(int)('I' - 'A')] =-0.075789256;	P[(int)('I' - 'A')][(int)('E' - 'A')] =-0.099708085;	P[(int)('I' - 'A')][(int)('N' - 'A')] =0.057789484;	P[(int)('I' - 'A')][(int)('T' - 'A')] =0.67627245;	P[(int)('I' - 'A')][(int)('D' - 'A')] =0.13906959;	P[(int)('I' - 'A')][(int)('V' - 'A')] =0.36882638;	P[(int)('I' - 'A')][(int)('A' - 'A')] =-0.018971558;	P[(int)('I' - 'A')][(int)('G' - 'A')] =-0.12231073;	P[(int)('I' - 'A')][(int)('R' - 'A')] =-0.29040155;	P[(int)('I' - 'A')][(int)('F' - 'A')] =0.13349877;	P[(int)('I' - 'A')][(int)('P' - 'A')] =-0.03418322;	P[(int)('I' - 'A')][(int)('Q' - 'A')] =0.10538742;	P[(int)('I' - 'A')][(int)('Y' - 'A')] =-0.036252559;	P[(int)('I' - 'A')][(int)('H' - 'A')] =-0.24891102;	P[(int)('I' - 'A')][(int)('M' - 'A')] =0.19198918;	P[(int)('I' - 'A')][(int)('C' - 'A')] =0.12826956;	P[(int)('I' - 'A')][(int)('W' - 'A')] =0.19037128;
		P[(int)('E' - 'A')][(int)('L' - 'A')] =-0.11527562;	P[(int)('E' - 'A')][(int)('S' - 'A')] =-0.22361229;	P[(int)('E' - 'A')][(int)('K' - 'A')] =0.020839499;	P[(int)('E' - 'A')][(int)('I' - 'A')] =0.39750661;	P[(int)('E' - 'A')][(int)('E' - 'A')] =-0.60082117;	P[(int)('E' - 'A')][(int)('N' - 'A')] =0.065694439;	P[(int)('E' - 'A')][(int)('T' - 'A')] =0.015931455;	P[(int)('E' - 'A')][(int)('D' - 'A')] =-0.19272853;	P[(int)('E' - 'A')][(int)('V' - 'A')] =-0.089954102;	P[(int)('E' - 'A')][(int)('A' - 'A')] =0.18966166;	P[(int)('E' - 'A')][(int)('G' - 'A')] =0.12248475;	P[(int)('E' - 'A')][(int)('R' - 'A')] =0.037424251;	P[(int)('E' - 'A')][(int)('F' - 'A')] =0.090412046;	P[(int)('E' - 'A')][(int)('P' - 'A')] =-0.092032984;	P[(int)('E' - 'A')][(int)('Q' - 'A')] =-0.29125188;	P[(int)('E' - 'A')][(int)('Y' - 'A')] =-0.31873233;	P[(int)('E' - 'A')][(int)('H' - 'A')] =-0.096089483;	P[(int)('E' - 'A')][(int)('M' - 'A')] =0.19509561;	P[(int)('E' - 'A')][(int)('C' - 'A')] =-0.19908013;	P[(int)('E' - 'A')][(int)('W' - 'A')] =-0.158567;
		P[(int)('N' - 'A')][(int)('L' - 'A')] =0.056280828;	P[(int)('N' - 'A')][(int)('S' - 'A')] =-0.2236092;	P[(int)('N' - 'A')][(int)('K' - 'A')] =-0.066768239;	P[(int)('N' - 'A')][(int)('I' - 'A')] =0.22575809;	P[(int)('N' - 'A')][(int)('E' - 'A')] =-0.16704824;	P[(int)('N' - 'A')][(int)('N' - 'A')] =0.055957086;	P[(int)('N' - 'A')][(int)('T' - 'A')] =-0.10093924;	P[(int)('N' - 'A')][(int)('D' - 'A')] =-0.17419701;	P[(int)('N' - 'A')][(int)('V' - 'A')] =-0.088212458;	P[(int)('N' - 'A')][(int)('A' - 'A')] =-0.21887226;	P[(int)('N' - 'A')][(int)('G' - 'A')] =-0.57601279;	P[(int)('N' - 'A')][(int)('R' - 'A')] =0.13139937;	P[(int)('N' - 'A')][(int)('F' - 'A')] =0.11968347;	P[(int)('N' - 'A')][(int)('P' - 'A')] =-0.052304864;	P[(int)('N' - 'A')][(int)('Q' - 'A')] =0.5153492;	P[(int)('N' - 'A')][(int)('Y' - 'A')] =-0.19490983;	P[(int)('N' - 'A')][(int)('H' - 'A')] =0.12052663;	P[(int)('N' - 'A')][(int)('M' - 'A')] =-0.23131157;	P[(int)('N' - 'A')][(int)('C' - 'A')] =0.092473946;	P[(int)('N' - 'A')][(int)('W' - 'A')] =0.13367521;
		P[(int)('T' - 'A')][(int)('L' - 'A')] =0.025933786;	P[(int)('T' - 'A')][(int)('S' - 'A')] =-0.22360866;	P[(int)('T' - 'A')][(int)('K' - 'A')] =-0.13824085;	P[(int)('T' - 'A')][(int)('I' - 'A')] =0.43823988;	P[(int)('T' - 'A')][(int)('E' - 'A')] =0.31454072;	P[(int)('T' - 'A')][(int)('N' - 'A')] =0.033740995;	P[(int)('T' - 'A')][(int)('T' - 'A')] =0.20580393;	P[(int)('T' - 'A')][(int)('D' - 'A')] =-0.094393202;	P[(int)('T' - 'A')][(int)('V' - 'A')] =0.031107812;	P[(int)('T' - 'A')][(int)('A' - 'A')] =0.014379639;	P[(int)('T' - 'A')][(int)('G' - 'A')] =0.26690077;	P[(int)('T' - 'A')][(int)('R' - 'A')] =-0.090428246;	P[(int)('T' - 'A')][(int)('F' - 'A')] =0.020309647;	P[(int)('T' - 'A')][(int)('P' - 'A')] =0.020872236;	P[(int)('T' - 'A')][(int)('Q' - 'A')] =-0.20406646;	P[(int)('T' - 'A')][(int)('Y' - 'A')] =-0.1157518;	P[(int)('T' - 'A')][(int)('H' - 'A')] =0.19337873;	P[(int)('T' - 'A')][(int)('M' - 'A')] =-0.40284698;	P[(int)('T' - 'A')][(int)('C' - 'A')] =0.49009384;	P[(int)('T' - 'A')][(int)('W' - 'A')] =-0.10224813;
		P[(int)('D' - 'A')][(int)('L' - 'A')] =0.013606676;	P[(int)('D' - 'A')][(int)('S' - 'A')] =-0.22361062;	P[(int)('D' - 'A')][(int)('K' - 'A')] =-0.004995016;	P[(int)('D' - 'A')][(int)('I' - 'A')] =-0.22920964;	P[(int)('D' - 'A')][(int)('E' - 'A')] =0.4481811;	P[(int)('D' - 'A')][(int)('N' - 'A')] =0.069961469;	P[(int)('D' - 'A')][(int)('T' - 'A')] =0.060847587;	P[(int)('D' - 'A')][(int)('D' - 'A')] =-0.22177809;	P[(int)('D' - 'A')][(int)('V' - 'A')] =-0.1260506;	P[(int)('D' - 'A')][(int)('A' - 'A')] =-0.077295497;	P[(int)('D' - 'A')][(int)('G' - 'A')] =0.30861569;	P[(int)('D' - 'A')][(int)('R' - 'A')] =0.1005807;	P[(int)('D' - 'A')][(int)('F' - 'A')] =0.12413203;	P[(int)('D' - 'A')][(int)('P' - 'A')] =-0.04336982;	P[(int)('D' - 'A')][(int)('Q' - 'A')] =0.24634832;	P[(int)('D' - 'A')][(int)('Y' - 'A')] =-0.56880075;	P[(int)('D' - 'A')][(int)('H' - 'A')] =-0.069512804;	P[(int)('D' - 'A')][(int)('M' - 'A')] =0.17794241;	P[(int)('D' - 'A')][(int)('C' - 'A')] =-0.2683052;	P[(int)('D' - 'A')][(int)('W' - 'A')] =-0.037296206;
		P[(int)('V' - 'A')][(int)('L' - 'A')] =-0.19863482;	P[(int)('V' - 'A')][(int)('S' - 'A')] =-0.22360883;	P[(int)('V' - 'A')][(int)('K' - 'A')] =-0.22800517;	P[(int)('V' - 'A')][(int)('I' - 'A')] =0.027314597;	P[(int)('V' - 'A')][(int)('E' - 'A')] =0.10511422;	P[(int)('V' - 'A')][(int)('N' - 'A')] =0.033630218;	P[(int)('V' - 'A')][(int)('T' - 'A')] =-0.59488338;	P[(int)('V' - 'A')][(int)('D' - 'A')] =0.077098026;	P[(int)('V' - 'A')][(int)('V' - 'A')] =0.29889248;	P[(int)('V' - 'A')][(int)('A' - 'A')] =0.0064628331;	P[(int)('V' - 'A')][(int)('G' - 'A')] =-0.01914737;	P[(int)('V' - 'A')][(int)('R' - 'A')] =-0.27264036;	P[(int)('V' - 'A')][(int)('F' - 'A')] =0.10426893;	P[(int)('V' - 'A')][(int)('P' - 'A')] =-0.0060224582;	P[(int)('V' - 'A')][(int)('Q' - 'A')] =-0.040577111;	P[(int)('V' - 'A')][(int)('Y' - 'A')] =-0.079907918;	P[(int)('V' - 'A')][(int)('H' - 'A')] =-0.41764962;	P[(int)('V' - 'A')][(int)('M' - 'A')] =0.18068771;	P[(int)('V' - 'A')][(int)('C' - 'A')] =0.26301386;	P[(int)('V' - 'A')][(int)('W' - 'A')] =0.16292375;
		P[(int)('A' - 'A')][(int)('L' - 'A')] =0.81006422;	P[(int)('A' - 'A')][(int)('S' - 'A')] =-0.22360703;	P[(int)('A' - 'A')][(int)('K' - 'A')] =-0.047940704;	P[(int)('A' - 'A')][(int)('I' - 'A')] =-0.33867696;	P[(int)('A' - 'A')][(int)('E' - 'A')] =-0.24786309;	P[(int)('A' - 'A')][(int)('N' - 'A')] =-0.010502185;	P[(int)('A' - 'A')][(int)('T' - 'A')] =-0.19260452;	P[(int)('A' - 'A')][(int)('D' - 'A')] =-0.095362209;	P[(int)('A' - 'A')][(int)('V' - 'A')] =0.0055011579;	P[(int)('A' - 'A')][(int)('A' - 'A')] =-0.0086609097;	P[(int)('A' - 'A')][(int)('G' - 'A')] =0.10024163;	P[(int)('A' - 'A')][(int)('R' - 'A')] =-0.077312896;	P[(int)('A' - 'A')][(int)('F' - 'A')] =-0.021612113;	P[(int)('A' - 'A')][(int)('P' - 'A')] =0.057106166;	P[(int)('A' - 'A')][(int)('Q' - 'A')] =-0.13057755;	P[(int)('A' - 'A')][(int)('Y' - 'A')] =-0.053051199;	P[(int)('A' - 'A')][(int)('H' - 'A')] =0.029966857;	P[(int)('A' - 'A')][(int)('M' - 'A')] =-0.085195542;	P[(int)('A' - 'A')][(int)('C' - 'A')] =0.13780212;	P[(int)('A' - 'A')][(int)('W' - 'A')] =-0.046719634;
		P[(int)('G' - 'A')][(int)('L' - 'A')] =-0.028080581;	P[(int)('G' - 'A')][(int)('S' - 'A')] =-0.22360932;	P[(int)('G' - 'A')][(int)('K' - 'A')] =-0.012213759;	P[(int)('G' - 'A')][(int)('I' - 'A')] =0.086038869;	P[(int)('G' - 'A')][(int)('E' - 'A')] =0.054801308;	P[(int)('G' - 'A')][(int)('N' - 'A')] =0.045041909;	P[(int)('G' - 'A')][(int)('T' - 'A')] =0.038073095;	P[(int)('G' - 'A')][(int)('D' - 'A')] =-0.16850832;	P[(int)('G' - 'A')][(int)('V' - 'A')] =-0.19339667;	P[(int)('G' - 'A')][(int)('A' - 'A')] =0.0076382216;	P[(int)('G' - 'A')][(int)('G' - 'A')] =-0.030815087;	P[(int)('G' - 'A')][(int)('R' - 'A')] =-0.09566834;	P[(int)('G' - 'A')][(int)('F' - 'A')] =0.13266633;	P[(int)('G' - 'A')][(int)('P' - 'A')] =0.86690294;	P[(int)('G' - 'A')][(int)('Q' - 'A')] =-0.0073609646;	P[(int)('G' - 'A')][(int)('Y' - 'A')] =0.23459406;	P[(int)('G' - 'A')][(int)('H' - 'A')] =-0.048627841;	P[(int)('G' - 'A')][(int)('M' - 'A')] =0.082929205;	P[(int)('G' - 'A')][(int)('C' - 'A')] =-0.14468318;	P[(int)('G' - 'A')][(int)('W' - 'A')] =0.069550548;
		P[(int)('R' - 'A')][(int)('L' - 'A')] =0.0012134753;	P[(int)('R' - 'A')][(int)('S' - 'A')] =-0.22360777;	P[(int)('R' - 'A')][(int)('K' - 'A')] =-0.023703424;	P[(int)('R' - 'A')][(int)('I' - 'A')] =0.09427355;	P[(int)('R' - 'A')][(int)('E' - 'A')] =-0.082673148;	P[(int)('R' - 'A')][(int)('N' - 'A')] =0.036629317;	P[(int)('R' - 'A')][(int)('T' - 'A')] =0.0043631193;	P[(int)('R' - 'A')][(int)('D' - 'A')] =-0.10006255;	P[(int)('R' - 'A')][(int)('V' - 'A')] =-0.12651398;	P[(int)('R' - 'A')][(int)('A' - 'A')] =-0.27850291;	P[(int)('R' - 'A')][(int)('G' - 'A')] =0.33205046;	P[(int)('R' - 'A')][(int)('R' - 'A')] =0.033884399;	P[(int)('R' - 'A')][(int)('F' - 'A')] =0.094577081;	P[(int)('R' - 'A')][(int)('P' - 'A')] =-0.22930682;	P[(int)('R' - 'A')][(int)('Q' - 'A')] =0.37835284;	P[(int)('R' - 'A')][(int)('Y' - 'A')] =0.53854561;	P[(int)('R' - 'A')][(int)('H' - 'A')] =-0.16521979;	P[(int)('R' - 'A')][(int)('M' - 'A')] =0.17677159;	P[(int)('R' - 'A')][(int)('C' - 'A')] =0.096943963;	P[(int)('R' - 'A')][(int)('W' - 'A')] =-0.39332143;
		P[(int)('F' - 'A')][(int)('L' - 'A')] =-0.016274645;	P[(int)('F' - 'A')][(int)('S' - 'A')] =-0.22360988;	P[(int)('F' - 'A')][(int)('K' - 'A')] =-0.1325802;	P[(int)('F' - 'A')][(int)('I' - 'A')] =-0.035525673;	P[(int)('F' - 'A')][(int)('E' - 'A')] =-0.021103564;	P[(int)('F' - 'A')][(int)('N' - 'A')] =0.072097367;	P[(int)('F' - 'A')][(int)('T' - 'A')] =0.054711969;	P[(int)('F' - 'A')][(int)('D' - 'A')] =0.37345021;	P[(int)('F' - 'A')][(int)('V' - 'A')] =0.24173567;	P[(int)('F' - 'A')][(int)('A' - 'A')] =0.0061805669;	P[(int)('F' - 'A')][(int)('G' - 'A')] =-0.014972304;	P[(int)('F' - 'A')][(int)('R' - 'A')] =0.39762628;	P[(int)('F' - 'A')][(int)('F' - 'A')] =-0.18511401;	P[(int)('F' - 'A')][(int)('P' - 'A')] =0.14717597;	P[(int)('F' - 'A')][(int)('Q' - 'A')] =0.0083443791;	P[(int)('F' - 'A')][(int)('Y' - 'A')] =-0.036615658;	P[(int)('F' - 'A')][(int)('H' - 'A')] =-0.40366501;	P[(int)('F' - 'A')][(int)('M' - 'A')] =-0.44739441;	P[(int)('F' - 'A')][(int)('C' - 'A')] =-0.25478508;	P[(int)('F' - 'A')][(int)('W' - 'A')] =-0.28393887;
		P[(int)('P' - 'A')][(int)('L' - 'A')] =-0.043331317;	P[(int)('P' - 'A')][(int)('S' - 'A')] =-0.22360965;	P[(int)('P' - 'A')][(int)('K' - 'A')] =-0.028801744;	P[(int)('P' - 'A')][(int)('I' - 'A')] =0.04830041;	P[(int)('P' - 'A')][(int)('E' - 'A')] =0.034746623;	P[(int)('P' - 'A')][(int)('N' - 'A')] =0.059993952;	P[(int)('P' - 'A')][(int)('T' - 'A')] =0.035027474;	P[(int)('P' - 'A')][(int)('D' - 'A')] =-0.20955872;	P[(int)('P' - 'A')][(int)('V' - 'A')] =-0.068930737;	P[(int)('P' - 'A')][(int)('A' - 'A')] =0.0013572205;	P[(int)('P' - 'A')][(int)('G' - 'A')] =-0.040223088;	P[(int)('P' - 'A')][(int)('R' - 'A')] =-0.25383376;	P[(int)('P' - 'A')][(int)('F' - 'A')] =-0.8886104;	P[(int)('P' - 'A')][(int)('P' - 'A')] =-0.061830409;	P[(int)('P' - 'A')][(int)('Q' - 'A')] =0.067189507;	P[(int)('P' - 'A')][(int)('Y' - 'A')] =0.036291502;	P[(int)('P' - 'A')][(int)('H' - 'A')] =-0.029071681;	P[(int)('P' - 'A')][(int)('M' - 'A')] =0.054341149;	P[(int)('P' - 'A')][(int)('C' - 'A')] =-0.098754421;	P[(int)('P' - 'A')][(int)('W' - 'A')] =0.10664713;
		P[(int)('Q' - 'A')][(int)('L' - 'A')] =-0.01707626;	P[(int)('Q' - 'A')][(int)('S' - 'A')] =-0.22360958;	P[(int)('Q' - 'A')][(int)('K' - 'A')] =0.0052537764;	P[(int)('Q' - 'A')][(int)('I' - 'A')] =-0.07965585;	P[(int)('Q' - 'A')][(int)('E' - 'A')] =0.23811778;	P[(int)('Q' - 'A')][(int)('N' - 'A')] =0.066159945;	P[(int)('Q' - 'A')][(int)('T' - 'A')] =0.05311656;	P[(int)('Q' - 'A')][(int)('D' - 'A')] =-0.16046821;	P[(int)('Q' - 'A')][(int)('V' - 'A')] =-0.074844139;	P[(int)('Q' - 'A')][(int)('A' - 'A')] =-0.47456179;	P[(int)('Q' - 'A')][(int)('G' - 'A')] =-0.38584913;	P[(int)('Q' - 'A')][(int)('R' - 'A')] =0.068146905;	P[(int)('Q' - 'A')][(int)('F' - 'A')] =0.098224246;	P[(int)('Q' - 'A')][(int)('P' - 'A')] =-0.20926201;	P[(int)('Q' - 'A')][(int)('Q' - 'A')] =-0.59931783;	P[(int)('Q' - 'A')][(int)('Y' - 'A')] =0.14736242;	P[(int)('Q' - 'A')][(int)('H' - 'A')] =-0.049119874;	P[(int)('Q' - 'A')][(int)('M' - 'A')] =0.061068559;	P[(int)('Q' - 'A')][(int)('C' - 'A')] =-0.15326971;	P[(int)('Q' - 'A')][(int)('W' - 'A')] =-0.073394469;
		P[(int)('Y' - 'A')][(int)('L' - 'A')] =0.011026047;	P[(int)('Y' - 'A')][(int)('S' - 'A')] =-0.22360971;	P[(int)('Y' - 'A')][(int)('K' - 'A')] =0.0036482368;	P[(int)('Y' - 'A')][(int)('I' - 'A')] =0.026420435;	P[(int)('Y' - 'A')][(int)('E' - 'A')] =0.024318876;	P[(int)('Y' - 'A')][(int)('N' - 'A')] =0.067788884;	P[(int)('Y' - 'A')][(int)('T' - 'A')] =-0.013127929;	P[(int)('Y' - 'A')][(int)('D' - 'A')] =0.24589231;	P[(int)('Y' - 'A')][(int)('V' - 'A')] =0.12930228;	P[(int)('Y' - 'A')][(int)('A' - 'A')] =-0.007384494;	P[(int)('Y' - 'A')][(int)('G' - 'A')] =-0.022245346;	P[(int)('Y' - 'A')][(int)('R' - 'A')] =0.57609145;	P[(int)('Y' - 'A')][(int)('F' - 'A')] =-0.18939991;	P[(int)('Y' - 'A')][(int)('P' - 'A')] =0.12654787;	P[(int)('Y' - 'A')][(int)('Q' - 'A')] =-0.041938123;	P[(int)('Y' - 'A')][(int)('Y' - 'A')] =-0.025821081;	P[(int)('Y' - 'A')][(int)('H' - 'A')] =0.31522837;	P[(int)('Y' - 'A')][(int)('M' - 'A')] =0.52674531;	P[(int)('Y' - 'A')][(int)('C' - 'A')] =0.31650149;	P[(int)('Y' - 'A')][(int)('W' - 'A')] =0.053574723;
		P[(int)('H' - 'A')][(int)('L' - 'A')] =-0.013280479;	P[(int)('H' - 'A')][(int)('S' - 'A')] =-0.22360981;	P[(int)('H' - 'A')][(int)('K' - 'A')] =0.0087193268;	P[(int)('H' - 'A')][(int)('I' - 'A')] =-0.025648073;	P[(int)('H' - 'A')][(int)('E' - 'A')] =-0.011632353;	P[(int)('H' - 'A')][(int)('N' - 'A')] =0.073091487;	P[(int)('H' - 'A')][(int)('T' - 'A')] =0.012962549;	P[(int)('H' - 'A')][(int)('D' - 'A')] =-0.1378315;	P[(int)('H' - 'A')][(int)('V' - 'A')] =-0.096282269;	P[(int)('H' - 'A')][(int)('A' - 'A')] =0.2186361;	P[(int)('H' - 'A')][(int)('G' - 'A')] =0.23335763;	P[(int)('H' - 'A')][(int)('R' - 'A')] =0.32198383;	P[(int)('H' - 'A')][(int)('F' - 'A')] =0.10819713;	P[(int)('H' - 'A')][(int)('P' - 'A')] =-0.22189768;	P[(int)('H' - 'A')][(int)('Q' - 'A')] =-0.051386707;	P[(int)('H' - 'A')][(int)('Y' - 'A')] =0.29577076;	P[(int)('H' - 'A')][(int)('H' - 'A')] =-0.10307492;	P[(int)('H' - 'A')][(int)('M' - 'A')] =-0.20170732;	P[(int)('H' - 'A')][(int)('C' - 'A')] =-0.098027687;	P[(int)('H' - 'A')][(int)('W' - 'A')] =0.7051635;
		P[(int)('M' - 'A')][(int)('L' - 'A')] =-0.11840265;	P[(int)('M' - 'A')][(int)('S' - 'A')] =-0.22360938;	P[(int)('M' - 'A')][(int)('K' - 'A')] =-0.54112533;	P[(int)('M' - 'A')][(int)('I' - 'A')] =-0.12008783;	P[(int)('M' - 'A')][(int)('E' - 'A')] =-0.085248353;	P[(int)('M' - 'A')][(int)('N' - 'A')] =0.058786293;	P[(int)('M' - 'A')][(int)('T' - 'A')] =0.02764003;	P[(int)('M' - 'A')][(int)('D' - 'A')] =0.11873405;	P[(int)('M' - 'A')][(int)('V' - 'A')] =0.28847802;	P[(int)('M' - 'A')][(int)('A' - 'A')] =0.015997304;	P[(int)('M' - 'A')][(int)('G' - 'A')] =0.069209881;	P[(int)('M' - 'A')][(int)('R' - 'A')] =-0.22135912;	P[(int)('M' - 'A')][(int)('F' - 'A')] =0.11831956;	P[(int)('M' - 'A')][(int)('P' - 'A')] =-0.067922872;	P[(int)('M' - 'A')][(int)('Q' - 'A')] =0.04040041;	P[(int)('M' - 'A')][(int)('Y' - 'A')] =0.11519895;	P[(int)('M' - 'A')][(int)('H' - 'A')] =0.5497466;	P[(int)('M' - 'A')][(int)('M' - 'A')] =-0.011140309;	P[(int)('M' - 'A')][(int)('C' - 'A')] =-0.36060996;	P[(int)('M' - 'A')][(int)('W' - 'A')] =-0.00087578136;
		P[(int)('C' - 'A')][(int)('L' - 'A')] =-0.033452759;	P[(int)('C' - 'A')][(int)('S' - 'A')] =-0.22356246;	P[(int)('C' - 'A')][(int)('K' - 'A')] =0.0023980888;	P[(int)('C' - 'A')][(int)('I' - 'A')] =0.026509773;	P[(int)('C' - 'A')][(int)('E' - 'A')] =0.020736643;	P[(int)('C' - 'A')][(int)('N' - 'A')] =-0.97083699;	P[(int)('C' - 'A')][(int)('T' - 'A')] =0.02414669;	P[(int)('C' - 'A')][(int)('D' - 'A')] =0.0058601905;	P[(int)('C' - 'A')][(int)('V' - 'A')] =0.0082559612;	P[(int)('C' - 'A')][(int)('A' - 'A')] =0.010497446;	P[(int)('C' - 'A')][(int)('G' - 'A')] =-0.01670019;	P[(int)('C' - 'A')][(int)('R' - 'A')] =0.034782281;	P[(int)('C' - 'A')][(int)('F' - 'A')] =-0.0067271241;	P[(int)('C' - 'A')][(int)('P' - 'A')] =-0.015134643;	P[(int)('C' - 'A')][(int)('Q' - 'A')] =0.0025445502;	P[(int)('C' - 'A')][(int)('Y' - 'A')] =-0.005672624;	P[(int)('C' - 'A')][(int)('H' - 'A')] =-0.003521271;	P[(int)('C' - 'A')][(int)('M' - 'A')] =0.015217687;	P[(int)('C' - 'A')][(int)('C' - 'A')] =-0.046367243;	P[(int)('C' - 'A')][(int)('W' - 'A')] =0.014812105;
		P[(int)('W' - 'A')][(int)('L' - 'A')] =0.0061977486;	P[(int)('W' - 'A')][(int)('S' - 'A')] =-0.22360924;	P[(int)('W' - 'A')][(int)('K' - 'A')] =-0.028610834;	P[(int)('W' - 'A')][(int)('I' - 'A')] =0.0010699833;	P[(int)('W' - 'A')][(int)('E' - 'A')] =0.0051139717;	P[(int)('W' - 'A')][(int)('N' - 'A')] =0.044264298;	P[(int)('W' - 'A')][(int)('T' - 'A')] =0.0086660113;	P[(int)('W' - 'A')][(int)('D' - 'A')] =0.65950168;	P[(int)('W' - 'A')][(int)('V' - 'A')] =-0.65919471;	P[(int)('W' - 'A')][(int)('A' - 'A')] =0.016016146;	P[(int)('W' - 'A')][(int)('G' - 'A')] =-0.025709822;	P[(int)('W' - 'A')][(int)('R' - 'A')] =-0.23255934;	P[(int)('W' - 'A')][(int)('F' - 'A')] =0.033317375;	P[(int)('W' - 'A')][(int)('P' - 'A')] =-0.0952961;	P[(int)('W' - 'A')][(int)('Q' - 'A')] =-0.019512703;	P[(int)('W' - 'A')][(int)('Y' - 'A')] =-0.074899145;	P[(int)('W' - 'A')][(int)('H' - 'A')] =0.0031703553;	P[(int)('W' - 'A')][(int)('M' - 'A')] =0.014112588;	P[(int)('W' - 'A')][(int)('C' - 'A')] =0.021192079;	P[(int)('W' - 'A')][(int)('W' - 'A')] =0.075915948;

		P1[(int)('L' - 'A')][(int)('L' - 'A')] =0.091052947;	P1[(int)('L' - 'A')][(int)('S' - 'A')] =-0.50203833;	P1[(int)('L' - 'A')][(int)('K' - 'A')] =-0.017277838;	P1[(int)('L' - 'A')][(int)('I' - 'A')] =0.087740304;	P1[(int)('L' - 'A')][(int)('E' - 'A')] =-0.11528022;	P1[(int)('L' - 'A')][(int)('N' - 'A')] =0.056280223;	P1[(int)('L' - 'A')][(int)('T' - 'A')] =0.025933455;	P1[(int)('L' - 'A')][(int)('D' - 'A')] =0.01360864;	P1[(int)('L' - 'A')][(int)('V' - 'A')] =-0.19863443;	P1[(int)('L' - 'A')][(int)('A' - 'A')] =0.81006313;	P1[(int)('L' - 'A')][(int)('G' - 'A')] =-0.028080363;	P1[(int)('L' - 'A')][(int)('R' - 'A')] =0.0012121415;	P1[(int)('L' - 'A')][(int)('F' - 'A')] =-0.016274461;	P1[(int)('L' - 'A')][(int)('P' - 'A')] =-0.043331061;	P1[(int)('L' - 'A')][(int)('Q' - 'A')] =-0.017075813;	P1[(int)('L' - 'A')][(int)('Y' - 'A')] =0.011026092;	P1[(int)('L' - 'A')][(int)('H' - 'A')] =-0.013280401;	P1[(int)('L' - 'A')][(int)('M' - 'A')] =-0.11840035;	P1[(int)('L' - 'A')][(int)('C' - 'A')] =-0.03344176;	P1[(int)('L' - 'A')][(int)('W' - 'A')] =0.0061977822;
		P1[(int)('S' - 'A')][(int)('L' - 'A')] =-0.2236076;	P1[(int)('S' - 'A')][(int)('S' - 'A')] =-0.22360063;	P1[(int)('S' - 'A')][(int)('K' - 'A')] =-0.22360696;	P1[(int)('S' - 'A')][(int)('I' - 'A')] =-0.22360828;	P1[(int)('S' - 'A')][(int)('E' - 'A')] =-0.2236072;	P1[(int)('S' - 'A')][(int)('N' - 'A')] =-0.22360658;	P1[(int)('S' - 'A')][(int)('T' - 'A')] =-0.22360083;	P1[(int)('S' - 'A')][(int)('D' - 'A')] =-0.22360722;	P1[(int)('S' - 'A')][(int)('V' - 'A')] =-0.22360752;	P1[(int)('S' - 'A')][(int)('A' - 'A')] =-0.22360593;	P1[(int)('S' - 'A')][(int)('G' - 'A')] =-0.22360679;	P1[(int)('S' - 'A')][(int)('R' - 'A')] =-0.22360752;	P1[(int)('S' - 'A')][(int)('F' - 'A')] =-0.2236086;	P1[(int)('S' - 'A')][(int)('P' - 'A')] =-0.22360701;	P1[(int)('S' - 'A')][(int)('Q' - 'A')] =-0.22360741;	P1[(int)('S' - 'A')][(int)('Y' - 'A')] =-0.22360823;	P1[(int)('S' - 'A')][(int)('H' - 'A')] =-0.22360793;	P1[(int)('S' - 'A')][(int)('M' - 'A')] =-0.22360618;	P1[(int)('S' - 'A')][(int)('C' - 'A')] =-0.22360878;	P1[(int)('S' - 'A')][(int)('W' - 'A')] =-0.22360876;
		P1[(int)('K' - 'A')][(int)('L' - 'A')] =0.70543711;	P1[(int)('K' - 'A')][(int)('S' - 'A')] =0.28278259;	P1[(int)('K' - 'A')][(int)('K' - 'A')] =0.056682896;	P1[(int)('K' - 'A')][(int)('I' - 'A')] =0.16732076;	P1[(int)('K' - 'A')][(int)('E' - 'A')] =0.020728881;	P1[(int)('K' - 'A')][(int)('N' - 'A')] =-0.066810565;	P1[(int)('K' - 'A')][(int)('T' - 'A')] =-0.1382724;	P1[(int)('K' - 'A')][(int)('D' - 'A')] =-0.0049305271;	P1[(int)('K' - 'A')][(int)('V' - 'A')] =-0.22799895;	P1[(int)('K' - 'A')][(int)('A' - 'A')] =-0.047922033;	P1[(int)('K' - 'A')][(int)('G' - 'A')] =-0.012219952;	P1[(int)('K' - 'A')][(int)('R' - 'A')] =-0.023729269;	P1[(int)('K' - 'A')][(int)('F' - 'A')] =-0.13257725;	P1[(int)('K' - 'A')][(int)('P' - 'A')] =-0.028804688;	P1[(int)('K' - 'A')][(int)('Q' - 'A')] =0.0052772337;	P1[(int)('K' - 'A')][(int)('Y' - 'A')] =0.0036468811;	P1[(int)('K' - 'A')][(int)('H' - 'A')] =0.0087233218;	P1[(int)('K' - 'A')][(int)('M' - 'A')] =-0.54111554;	P1[(int)('K' - 'A')][(int)('C' - 'A')] =0.0023924803;	P1[(int)('K' - 'A')][(int)('W' - 'A')] =-0.028610265;
		P1[(int)('I' - 'A')][(int)('L' - 'A')] =0.23874568;	P1[(int)('I' - 'A')][(int)('S' - 'A')] =-0.54846086;	P1[(int)('I' - 'A')][(int)('K' - 'A')] =-0.15726536;	P1[(int)('I' - 'A')][(int)('I' - 'A')] =-0.07581826;	P1[(int)('I' - 'A')][(int)('E' - 'A')] =0.39746364;	P1[(int)('I' - 'A')][(int)('N' - 'A')] =0.22575718;	P1[(int)('I' - 'A')][(int)('T' - 'A')] =0.43828016;	P1[(int)('I' - 'A')][(int)('D' - 'A')] =-0.22918148;	P1[(int)('I' - 'A')][(int)('V' - 'A')] =0.027356669;	P1[(int)('I' - 'A')][(int)('A' - 'A')] =-0.33868852;	P1[(int)('I' - 'A')][(int)('G' - 'A')] =0.086044174;	P1[(int)('I' - 'A')][(int)('R' - 'A')] =0.094270009;	P1[(int)('I' - 'A')][(int)('F' - 'A')] =-0.035506623;	P1[(int)('I' - 'A')][(int)('P' - 'A')] =0.048306904;	P1[(int)('I' - 'A')][(int)('Q' - 'A')] =-0.079642271;	P1[(int)('I' - 'A')][(int)('Y' - 'A')] =0.026420875;	P1[(int)('I' - 'A')][(int)('H' - 'A')] =-0.025649642;	P1[(int)('I' - 'A')][(int)('M' - 'A')] =-0.1200289;	P1[(int)('I' - 'A')][(int)('C' - 'A')] =0.026516857;	P1[(int)('I' - 'A')][(int)('W' - 'A')] =0.0010745368;
		P1[(int)('E' - 'A')][(int)('L' - 'A')] =0.13377431;	P1[(int)('E' - 'A')][(int)('S' - 'A')] =-0.28890377;	P1[(int)('E' - 'A')][(int)('K' - 'A')] =0.22563252;	P1[(int)('E' - 'A')][(int)('I' - 'A')] =-0.099696861;	P1[(int)('E' - 'A')][(int)('E' - 'A')] =-0.6007933;	P1[(int)('E' - 'A')][(int)('N' - 'A')] =-0.16703957;	P1[(int)('E' - 'A')][(int)('T' - 'A')] =0.31455888;	P1[(int)('E' - 'A')][(int)('D' - 'A')] =0.44816431;	P1[(int)('E' - 'A')][(int)('V' - 'A')] =0.10508517;	P1[(int)('E' - 'A')][(int)('A' - 'A')] =-0.2478863;	P1[(int)('E' - 'A')][(int)('G' - 'A')] =0.054805799;	P1[(int)('E' - 'A')][(int)('R' - 'A')] =-0.082671051;	P1[(int)('E' - 'A')][(int)('F' - 'A')] =-0.021122172;	P1[(int)('E' - 'A')][(int)('P' - 'A')] =0.034746558;	P1[(int)('E' - 'A')][(int)('Q' - 'A')] =0.23810911;	P1[(int)('E' - 'A')][(int)('Y' - 'A')] =0.024321902;	P1[(int)('E' - 'A')][(int)('H' - 'A')] =-0.011632656;	P1[(int)('E' - 'A')][(int)('M' - 'A')] =-0.085293473;	P1[(int)('E' - 'A')][(int)('C' - 'A')] =0.02073991;	P1[(int)('E' - 'A')][(int)('W' - 'A')] =0.0051109242;
		P1[(int)('N' - 'A')][(int)('L' - 'A')] =0.05971272;	P1[(int)('N' - 'A')][(int)('S' - 'A')] =0.022229596;	P1[(int)('N' - 'A')][(int)('K' - 'A')] =0.058775924;	P1[(int)('N' - 'A')][(int)('I' - 'A')] =0.057758784;	P1[(int)('N' - 'A')][(int)('E' - 'A')] =0.06569062;	P1[(int)('N' - 'A')][(int)('N' - 'A')] =0.055958238;	P1[(int)('N' - 'A')][(int)('T' - 'A')] =0.033772145;	P1[(int)('N' - 'A')][(int)('D' - 'A')] =0.069959937;	P1[(int)('N' - 'A')][(int)('V' - 'A')] =0.033609307;	P1[(int)('N' - 'A')][(int)('A' - 'A')] =-0.010500854;	P1[(int)('N' - 'A')][(int)('G' - 'A')] =0.045043367;	P1[(int)('N' - 'A')][(int)('R' - 'A')] =0.03662013;	P1[(int)('N' - 'A')][(int)('F' - 'A')] =0.072056019;	P1[(int)('N' - 'A')][(int)('P' - 'A')] =0.059993926;	P1[(int)('N' - 'A')][(int)('Q' - 'A')] =0.066152844;	P1[(int)('N' - 'A')][(int)('Y' - 'A')] =0.067757635;	P1[(int)('N' - 'A')][(int)('H' - 'A')] =0.073079786;	P1[(int)('N' - 'A')][(int)('M' - 'A')] =0.058758747;	P1[(int)('N' - 'A')][(int)('C' - 'A')] =-0.97085072;	P1[(int)('N' - 'A')][(int)('W' - 'A')] =0.044217921;
		P1[(int)('T' - 'A')][(int)('L' - 'A')] =-0.28980515;	P1[(int)('T' - 'A')][(int)('S' - 'A')] =-0.02997398;	P1[(int)('T' - 'A')][(int)('K' - 'A')] =0.0037668537;	P1[(int)('T' - 'A')][(int)('I' - 'A')] =0.67627156;	P1[(int)('T' - 'A')][(int)('E' - 'A')] =0.015928323;	P1[(int)('T' - 'A')][(int)('N' - 'A')] =-0.10093981;	P1[(int)('T' - 'A')][(int)('T' - 'A')] =0.20580713;	P1[(int)('T' - 'A')][(int)('D' - 'A')] =0.060850379;	P1[(int)('T' - 'A')][(int)('V' - 'A')] =-0.59488316;	P1[(int)('T' - 'A')][(int)('A' - 'A')] =-0.19260365;	P1[(int)('T' - 'A')][(int)('G' - 'A')] =0.038072939;	P1[(int)('T' - 'A')][(int)('R' - 'A')] =0.0043653335;	P1[(int)('T' - 'A')][(int)('F' - 'A')] =0.054712005;	P1[(int)('T' - 'A')][(int)('P' - 'A')] =0.035027275;	P1[(int)('T' - 'A')][(int)('Q' - 'A')] =0.053120335;	P1[(int)('T' - 'A')][(int)('Y' - 'A')] =-0.013127919;	P1[(int)('T' - 'A')][(int)('H' - 'A')] =0.012961442;	P1[(int)('T' - 'A')][(int)('M' - 'A')] =0.027639974;	P1[(int)('T' - 'A')][(int)('C' - 'A')] =0.024145111;	P1[(int)('T' - 'A')][(int)('W' - 'A')] =0.0086659092;
		P1[(int)('D' - 'A')][(int)('L' - 'A')] =0.18079849;	P1[(int)('D' - 'A')][(int)('S' - 'A')] =-0.098979062;	P1[(int)('D' - 'A')][(int)('K' - 'A')] =-0.14655603;	P1[(int)('D' - 'A')][(int)('I' - 'A')] =0.13906719;	P1[(int)('D' - 'A')][(int)('E' - 'A')] =-0.1927348;	P1[(int)('D' - 'A')][(int)('N' - 'A')] =-0.17420058;	P1[(int)('D' - 'A')][(int)('T' - 'A')] =-0.09439458;	P1[(int)('D' - 'A')][(int)('D' - 'A')] =-0.22178391;	P1[(int)('D' - 'A')][(int)('V' - 'A')] =0.077097199;	P1[(int)('D' - 'A')][(int)('A' - 'A')] =-0.095360464;	P1[(int)('D' - 'A')][(int)('G' - 'A')] =-0.16851095;	P1[(int)('D' - 'A')][(int)('R' - 'A')] =-0.10006316;	P1[(int)('D' - 'A')][(int)('F' - 'A')] =0.37344495;	P1[(int)('D' - 'A')][(int)('P' - 'A')] =-0.20956253;	P1[(int)('D' - 'A')][(int)('Q' - 'A')] =-0.16047226;	P1[(int)('D' - 'A')][(int)('Y' - 'A')] =0.24588694;	P1[(int)('D' - 'A')][(int)('H' - 'A')] =-0.13783678;	P1[(int)('D' - 'A')][(int)('M' - 'A')] =0.11873222;	P1[(int)('D' - 'A')][(int)('C' - 'A')] =0.0059285985;	P1[(int)('D' - 'A')][(int)('W' - 'A')] =0.65949886;
		P1[(int)('V' - 'A')][(int)('L' - 'A')] =0.26339172;	P1[(int)('V' - 'A')][(int)('S' - 'A')] =-0.04608943;	P1[(int)('V' - 'A')][(int)('K' - 'A')] =-0.066014585;	P1[(int)('V' - 'A')][(int)('I' - 'A')] =0.36882631;	P1[(int)('V' - 'A')][(int)('E' - 'A')] =-0.089961017;	P1[(int)('V' - 'A')][(int)('N' - 'A')] =-0.088213783;	P1[(int)('V' - 'A')][(int)('T' - 'A')] =0.031107316;	P1[(int)('V' - 'A')][(int)('D' - 'A')] =-0.12605556;	P1[(int)('V' - 'A')][(int)('V' - 'A')] =0.29889232;	P1[(int)('V' - 'A')][(int)('A' - 'A')] =0.0055009393;	P1[(int)('V' - 'A')][(int)('G' - 'A')] =-0.19339867;	P1[(int)('V' - 'A')][(int)('R' - 'A')] =-0.12651054;	P1[(int)('V' - 'A')][(int)('F' - 'A')] =0.2417334;	P1[(int)('V' - 'A')][(int)('P' - 'A')] =-0.068932722;	P1[(int)('V' - 'A')][(int)('Q' - 'A')] =-0.074844764;	P1[(int)('V' - 'A')][(int)('Y' - 'A')] =0.12929945;	P1[(int)('V' - 'A')][(int)('H' - 'A')] =-0.096284049;	P1[(int)('V' - 'A')][(int)('M' - 'A')] =0.28847872;	P1[(int)('V' - 'A')][(int)('C' - 'A')] =0.0082696763;	P1[(int)('V' - 'A')][(int)('W' - 'A')] =-0.65919406;
		P1[(int)('A' - 'A')][(int)('L' - 'A')] =-0.025494345;	P1[(int)('A' - 'A')][(int)('S' - 'A')] =-0.11480157;	P1[(int)('A' - 'A')][(int)('K' - 'A')] =0.73775397;	P1[(int)('A' - 'A')][(int)('I' - 'A')] =-0.018978851;	P1[(int)('A' - 'A')][(int)('E' - 'A')] =0.18965979;	P1[(int)('A' - 'A')][(int)('N' - 'A')] =-0.21887689;	P1[(int)('A' - 'A')][(int)('T' - 'A')] =0.01438189;	P1[(int)('A' - 'A')][(int)('D' - 'A')] =-0.077291224;	P1[(int)('A' - 'A')][(int)('V' - 'A')] =0.0064506842;	P1[(int)('A' - 'A')][(int)('A' - 'A')] =-0.0086606862;	P1[(int)('A' - 'A')][(int)('G' - 'A')] =0.0076379301;	P1[(int)('A' - 'A')][(int)('R' - 'A')] =-0.27850299;	P1[(int)('A' - 'A')][(int)('F' - 'A')] =0.0061690873;	P1[(int)('A' - 'A')][(int)('P' - 'A')] =0.0013556533;	P1[(int)('A' - 'A')][(int)('Q' - 'A')] =-0.47456558;	P1[(int)('A' - 'A')][(int)('Y' - 'A')] =-0.0073817388;	P1[(int)('A' - 'A')][(int)('H' - 'A')] =0.21863721;	P1[(int)('A' - 'A')][(int)('M' - 'A')] =0.01600382;	P1[(int)('A' - 'A')][(int)('C' - 'A')] =0.010500823;	P1[(int)('A' - 'A')][(int)('W' - 'A')] =0.016014412;
		P1[(int)('G' - 'A')][(int)('L' - 'A')] =0.072451966;	P1[(int)('G' - 'A')][(int)('S' - 'A')] =0.10005098;	P1[(int)('G' - 'A')][(int)('K' - 'A')] =-0.35138834;	P1[(int)('G' - 'A')][(int)('I' - 'A')] =-0.12230544;	P1[(int)('G' - 'A')][(int)('E' - 'A')] =0.12248824;	P1[(int)('G' - 'A')][(int)('N' - 'A')] =-0.57602091;	P1[(int)('G' - 'A')][(int)('T' - 'A')] =0.26690126;	P1[(int)('G' - 'A')][(int)('D' - 'A')] =0.30861723;	P1[(int)('G' - 'A')][(int)('V' - 'A')] =-0.019132883;	P1[(int)('G' - 'A')][(int)('A' - 'A')] =0.10024106;	P1[(int)('G' - 'A')][(int)('G' - 'A')] =-0.030814406;	P1[(int)('G' - 'A')][(int)('R' - 'A')] =0.33204824;	P1[(int)('G' - 'A')][(int)('F' - 'A')] =-0.014962296;	P1[(int)('G' - 'A')][(int)('P' - 'A')] =-0.040222151;	P1[(int)('G' - 'A')][(int)('Q' - 'A')] =-0.38585102;	P1[(int)('G' - 'A')][(int)('Y' - 'A')] =-0.022248715;	P1[(int)('G' - 'A')][(int)('H' - 'A')] =0.23336229;	P1[(int)('G' - 'A')][(int)('M' - 'A')] =0.069195592;	P1[(int)('G' - 'A')][(int)('C' - 'A')] =-0.0167115;	P1[(int)('G' - 'A')][(int)('W' - 'A')] =-0.025708327;
		P1[(int)('R' - 'A')][(int)('L' - 'A')] =-0.14277283;	P1[(int)('R' - 'A')][(int)('S' - 'A')] =-0.026085099;	P1[(int)('R' - 'A')][(int)('K' - 'A')] =0.0011250205;	P1[(int)('R' - 'A')][(int)('I' - 'A')] =-0.29040251;	P1[(int)('R' - 'A')][(int)('E' - 'A')] =0.037430349;	P1[(int)('R' - 'A')][(int)('N' - 'A')] =0.1314001;	P1[(int)('R' - 'A')][(int)('T' - 'A')] =-0.090430069;	P1[(int)('R' - 'A')][(int)('D' - 'A')] =0.10058555;	P1[(int)('R' - 'A')][(int)('V' - 'A')] =-0.27264074;	P1[(int)('R' - 'A')][(int)('A' - 'A')] =-0.077311837;	P1[(int)('R' - 'A')][(int)('G' - 'A')] =-0.095664813;	P1[(int)('R' - 'A')][(int)('R' - 'A')] =0.033881166;	P1[(int)('R' - 'A')][(int)('F' - 'A')] =0.39762504;	P1[(int)('R' - 'A')][(int)('P' - 'A')] =-0.25383126;	P1[(int)('R' - 'A')][(int)('Q' - 'A')] =0.068147075;	P1[(int)('R' - 'A')][(int)('Y' - 'A')] =0.57609088;	P1[(int)('R' - 'A')][(int)('H' - 'A')] =0.32198455;	P1[(int)('R' - 'A')][(int)('M' - 'A')] =-0.22135969;	P1[(int)('R' - 'A')][(int)('C' - 'A')] =0.034790502;	P1[(int)('R' - 'A')][(int)('W' - 'A')] =-0.23255861;
		P1[(int)('F' - 'A')][(int)('L' - 'A')] =0.03787369;	P1[(int)('F' - 'A')][(int)('S' - 'A')] =-0.013850912;	P1[(int)('F' - 'A')][(int)('K' - 'A')] =0.089840537;	P1[(int)('F' - 'A')][(int)('I' - 'A')] =0.13349994;	P1[(int)('F' - 'A')][(int)('E' - 'A')] =0.090408418;	P1[(int)('F' - 'A')][(int)('N' - 'A')] =0.11968276;	P1[(int)('F' - 'A')][(int)('T' - 'A')] =0.020310491;	P1[(int)('F' - 'A')][(int)('D' - 'A')] =0.12412861;	P1[(int)('F' - 'A')][(int)('V' - 'A')] =0.10426997;	P1[(int)('F' - 'A')][(int)('A' - 'A')] =-0.02161225;	P1[(int)('F' - 'A')][(int)('G' - 'A')] =0.13266302;	P1[(int)('F' - 'A')][(int)('R' - 'A')] =0.094580019;	P1[(int)('F' - 'A')][(int)('F' - 'A')] =-0.1851142;	P1[(int)('F' - 'A')][(int)('P' - 'A')] =-0.88861057;	P1[(int)('F' - 'A')][(int)('Q' - 'A')] =0.09822445;	P1[(int)('F' - 'A')][(int)('Y' - 'A')] =-0.18940069;	P1[(int)('F' - 'A')][(int)('H' - 'A')] =0.10819656;	P1[(int)('F' - 'A')][(int)('M' - 'A')] =0.11831955;	P1[(int)('F' - 'A')][(int)('C' - 'A')] =-0.0067257086;	P1[(int)('F' - 'A')][(int)('W' - 'A')] =0.033316959;
		P1[(int)('P' - 'A')][(int)('L' - 'A')] =-0.031141747;	P1[(int)('P' - 'A')][(int)('S' - 'A')] =0.086211533;	P1[(int)('P' - 'A')][(int)('K' - 'A')] =-0.14510971;	P1[(int)('P' - 'A')][(int)('I' - 'A')] =-0.034186972;	P1[(int)('P' - 'A')][(int)('E' - 'A')] =-0.092030502;	P1[(int)('P' - 'A')][(int)('N' - 'A')] =-0.052304121;	P1[(int)('P' - 'A')][(int)('T' - 'A')] =0.020871572;	P1[(int)('P' - 'A')][(int)('D' - 'A')] =-0.043367165;	P1[(int)('P' - 'A')][(int)('V' - 'A')] =-0.0060265183;	P1[(int)('P' - 'A')][(int)('A' - 'A')] =0.057105183;	P1[(int)('P' - 'A')][(int)('G' - 'A')] =0.8669025;	P1[(int)('P' - 'A')][(int)('R' - 'A')] =-0.22930893;	P1[(int)('P' - 'A')][(int)('F' - 'A')] =0.14717678;	P1[(int)('P' - 'A')][(int)('P' - 'A')] =-0.061827211;	P1[(int)('P' - 'A')][(int)('Q' - 'A')] =-0.20926163;	P1[(int)('P' - 'A')][(int)('Y' - 'A')] =0.12654997;	P1[(int)('P' - 'A')][(int)('H' - 'A')] =-0.221896;	P1[(int)('P' - 'A')][(int)('M' - 'A')] =-0.067921803;	P1[(int)('P' - 'A')][(int)('C' - 'A')] =-0.015144733;	P1[(int)('P' - 'A')][(int)('W' - 'A')] =-0.095295184;
		P1[(int)('Q' - 'A')][(int)('L' - 'A')] =0.039746857;	P1[(int)('Q' - 'A')][(int)('S' - 'A')] =-0.042956533;	P1[(int)('Q' - 'A')][(int)('K' - 'A')] =0.025296168;	P1[(int)('Q' - 'A')][(int)('I' - 'A')] =0.10538133;	P1[(int)('Q' - 'A')][(int)('E' - 'A')] =-0.29125463;	P1[(int)('Q' - 'A')][(int)('N' - 'A')] =0.51535379;	P1[(int)('Q' - 'A')][(int)('T' - 'A')] =-0.20406788;	P1[(int)('Q' - 'A')][(int)('D' - 'A')] =0.24634729;	P1[(int)('Q' - 'A')][(int)('V' - 'A')] =-0.040590963;	P1[(int)('Q' - 'A')][(int)('A' - 'A')] =-0.13057786;	P1[(int)('Q' - 'A')][(int)('G' - 'A')] =-0.0073611857;	P1[(int)('Q' - 'A')][(int)('R' - 'A')] =0.37834416;	P1[(int)('Q' - 'A')][(int)('F' - 'A')] =0.0083336593;	P1[(int)('Q' - 'A')][(int)('P' - 'A')] =0.067189229;	P1[(int)('Q' - 'A')][(int)('Q' - 'A')] =-0.59931602;	P1[(int)('Q' - 'A')][(int)('Y' - 'A')] =-0.041932131;	P1[(int)('Q' - 'A')][(int)('H' - 'A')] =-0.051389359;	P1[(int)('Q' - 'A')][(int)('M' - 'A')] =0.040418367;	P1[(int)('Q' - 'A')][(int)('C' - 'A')] =0.0025534551;	P1[(int)('Q' - 'A')][(int)('W' - 'A')] =-0.019513026;
		P1[(int)('Y' - 'A')][(int)('L' - 'A')] =0.059897546;	P1[(int)('Y' - 'A')][(int)('S' - 'A')] =-0.079993517;	P1[(int)('Y' - 'A')][(int)('K' - 'A')] =0.16275292;	P1[(int)('Y' - 'A')][(int)('I' - 'A')] =-0.036238813;	P1[(int)('Y' - 'A')][(int)('E' - 'A')] =-0.31873875;	P1[(int)('Y' - 'A')][(int)('N' - 'A')] =-0.19490997;	P1[(int)('Y' - 'A')][(int)('T' - 'A')] =-0.11573723;	P1[(int)('Y' - 'A')][(int)('D' - 'A')] =-0.5688097;	P1[(int)('Y' - 'A')][(int)('V' - 'A')] =-0.079888561;	P1[(int)('Y' - 'A')][(int)('A' - 'A')] =-0.053045837;	P1[(int)('Y' - 'A')][(int)('G' - 'A')] =0.23458651;	P1[(int)('Y' - 'A')][(int)('R' - 'A')] =0.53855346;	P1[(int)('Y' - 'A')][(int)('F' - 'A')] =-0.036615059;	P1[(int)('Y' - 'A')][(int)('P' - 'A')] =0.036284543;	P1[(int)('Y' - 'A')][(int)('Q' - 'A')] =0.14735905;	P1[(int)('Y' - 'A')][(int)('Y' - 'A')] =-0.025820564;	P1[(int)('Y' - 'A')][(int)('H' - 'A')] =0.29576062;	P1[(int)('Y' - 'A')][(int)('M' - 'A')] =0.11517783;	P1[(int)('Y' - 'A')][(int)('C' - 'A')] =-0.0056636663;	P1[(int)('Y' - 'A')][(int)('W' - 'A')] =-0.074900046;
		P1[(int)('H' - 'A')][(int)('L' - 'A')] =0.25149574;	P1[(int)('H' - 'A')][(int)('S' - 'A')] =0.15880415;	P1[(int)('H' - 'A')][(int)('K' - 'A')] =0.012162569;	P1[(int)('H' - 'A')][(int)('I' - 'A')] =-0.24889263;	P1[(int)('H' - 'A')][(int)('E' - 'A')] =-0.096092054;	P1[(int)('H' - 'A')][(int)('N' - 'A')] =0.12053879;	P1[(int)('H' - 'A')][(int)('T' - 'A')] =0.19333418;	P1[(int)('H' - 'A')][(int)('D' - 'A')] =-0.069481829;	P1[(int)('H' - 'A')][(int)('V' - 'A')] =-0.41764078;	P1[(int)('H' - 'A')][(int)('A' - 'A')] =0.029951008;	P1[(int)('H' - 'A')][(int)('G' - 'A')] =-0.048620127;	P1[(int)('H' - 'A')][(int)('R' - 'A')] =-0.16523793;	P1[(int)('H' - 'A')][(int)('F' - 'A')] =-0.40368678;	P1[(int)('H' - 'A')][(int)('P' - 'A')] =-0.029061843;	P1[(int)('H' - 'A')][(int)('Q' - 'A')] =-0.04913248;	P1[(int)('H' - 'A')][(int)('Y' - 'A')] =0.3152514;	P1[(int)('H' - 'A')][(int)('H' - 'A')] =-0.10308001;	P1[(int)('H' - 'A')][(int)('M' - 'A')] =0.54975002;	P1[(int)('H' - 'A')][(int)('C' - 'A')] =-0.0035335925;	P1[(int)('H' - 'A')][(int)('W' - 'A')] =0.0031737286;
		P1[(int)('M' - 'A')][(int)('L' - 'A')] =-0.082054143;	P1[(int)('M' - 'A')][(int)('S' - 'A')] =-0.25618577;	P1[(int)('M' - 'A')][(int)('K' - 'A')] =0.04093299;	P1[(int)('M' - 'A')][(int)('I' - 'A')] =0.19197339;	P1[(int)('M' - 'A')][(int)('E' - 'A')] =0.19508986;	P1[(int)('M' - 'A')][(int)('N' - 'A')] =-0.23130485;	P1[(int)('M' - 'A')][(int)('T' - 'A')] =-0.40283624;	P1[(int)('M' - 'A')][(int)('D' - 'A')] =0.17793879;	P1[(int)('M' - 'A')][(int)('V' - 'A')] =0.18066551;	P1[(int)('M' - 'A')][(int)('A' - 'A')] =-0.08519085;	P1[(int)('M' - 'A')][(int)('G' - 'A')] =0.082925067;	P1[(int)('M' - 'A')][(int)('R' - 'A')] =0.17676293;	P1[(int)('M' - 'A')][(int)('F' - 'A')] =-0.44741708;	P1[(int)('M' - 'A')][(int)('P' - 'A')] =0.054338998;	P1[(int)('M' - 'A')][(int)('Q' - 'A')] =0.061063675;	P1[(int)('M' - 'A')][(int)('Y' - 'A')] =0.52676335;	P1[(int)('M' - 'A')][(int)('H' - 'A')] =-0.20171275;	P1[(int)('M' - 'A')][(int)('M' - 'A')] =-0.011111369;	P1[(int)('M' - 'A')][(int)('C' - 'A')] =0.015244925;	P1[(int)('M' - 'A')][(int)('W' - 'A')] =0.014112459;
		P1[(int)('C' - 'A')][(int)('L' - 'A')] =-0.27492262;	P1[(int)('C' - 'A')][(int)('S' - 'A')] =0.2911152;	P1[(int)('C' - 'A')][(int)('K' - 'A')] =0.061431947;	P1[(int)('C' - 'A')][(int)('I' - 'A')] =0.12826771;	P1[(int)('C' - 'A')][(int)('E' - 'A')] =-0.19907014;	P1[(int)('C' - 'A')][(int)('N' - 'A')] =0.092451265;	P1[(int)('C' - 'A')][(int)('T' - 'A')] =0.49009606;	P1[(int)('C' - 'A')][(int)('D' - 'A')] =-0.2683228;	P1[(int)('C' - 'A')][(int)('V' - 'A')] =0.26301352;	P1[(int)('C' - 'A')][(int)('A' - 'A')] =0.13780113;	P1[(int)('C' - 'A')][(int)('G' - 'A')] =-0.14467953;	P1[(int)('C' - 'A')][(int)('R' - 'A')] =0.09699058;	P1[(int)('C' - 'A')][(int)('F' - 'A')] =-0.25476527;	P1[(int)('C' - 'A')][(int)('P' - 'A')] =-0.098759419;	P1[(int)('C' - 'A')][(int)('Q' - 'A')] =-0.15325216;	P1[(int)('C' - 'A')][(int)('Y' - 'A')] =0.31649031;	P1[(int)('C' - 'A')][(int)('H' - 'A')] =-0.09805021;	P1[(int)('C' - 'A')][(int)('M' - 'A')] =-0.36061588;	P1[(int)('C' - 'A')][(int)('C' - 'A')] =-0.046404199;	P1[(int)('C' - 'A')][(int)('W' - 'A')] =0.021186457;
		P1[(int)('W' - 'A')][(int)('L' - 'A')] =-0.010229702;	P1[(int)('W' - 'A')][(int)('S' - 'A')] =-0.071490404;	P1[(int)('W' - 'A')][(int)('K' - 'A')] =-0.33456422;	P1[(int)('W' - 'A')][(int)('I' - 'A')] =0.19035667;	P1[(int)('W' - 'A')][(int)('E' - 'A')] =-0.15855578;	P1[(int)('W' - 'A')][(int)('N' - 'A')] =0.13367485;	P1[(int)('W' - 'A')][(int)('T' - 'A')] =-0.10226951;	P1[(int)('W' - 'A')][(int)('D' - 'A')] =-0.037277677;	P1[(int)('W' - 'A')][(int)('V' - 'A')] =0.16290187;	P1[(int)('W' - 'A')][(int)('A' - 'A')] =-0.046725004;	P1[(int)('W' - 'A')][(int)('G' - 'A')] =0.069555789;	P1[(int)('W' - 'A')][(int)('R' - 'A')] =-0.39333354;	P1[(int)('W' - 'A')][(int)('F' - 'A')] =-0.28393251;	P1[(int)('W' - 'A')][(int)('P' - 'A')] =0.10665249;	P1[(int)('W' - 'A')][(int)('Q' - 'A')] =-0.07339062;	P1[(int)('W' - 'A')][(int)('Y' - 'A')] =0.05356669;	P1[(int)('W' - 'A')][(int)('H' - 'A')] =0.70516439;	P1[(int)('W' - 'A')][(int)('M' - 'A')] =-0.00084960969;	P1[(int)('W' - 'A')][(int)('C' - 'A')] =0.014824737;	P1[(int)('W' - 'A')][(int)('W' - 'A')] =0.075915656;
	
		sum_letters[(int)('A' - 'A')] = 0.0866279;
		sum_letters[(int)('R' - 'A')] = 0.043972;
		sum_letters[(int)('N' - 'A')] = 0.0390894;
		sum_letters[(int)('D' - 'A')] = 0.0570451;
		sum_letters[(int)('C' - 'A')] = 0.0193078;
		sum_letters[(int)('Q' - 'A')] = 0.0367281;
		sum_letters[(int)('E' - 'A')] = 0.0580589;
		sum_letters[(int)('G' - 'A')] = 0.0832518;
		sum_letters[(int)('H' - 'A')] = 0.0244313;
		sum_letters[(int)('I' - 'A')] = 0.048466;
		sum_letters[(int)('L' - 'A')] = 0.086209;
		sum_letters[(int)('K' - 'A')] = 0.0620286;
		sum_letters[(int)('M' - 'A')] = 0.0195027;
		sum_letters[(int)('F' - 'A')] = 0.0384319;
		sum_letters[(int)('P' - 'A')] = 0.0457631;
		sum_letters[(int)('S' - 'A')] = 0.0695179;
		sum_letters[(int)('T' - 'A')] = 0.0610127;
		sum_letters[(int)('W' - 'A')] = 0.0143859;
		sum_letters[(int)('Y' - 'A')] = 0.0352742;
		sum_letters[(int)('V' - 'A')] = 0.0708956;
	}
	
	
	

	//######################################################################
	// Populate substitution matrix for the gap model
	// We will hardcode the matrix instead.
	//######################################################################
	
	lambda_gap[((int)('X' - 88)/-43)] = -0.002402885;
	lambda_gap[((int)('-' - 88)/-43)] = 0;
	
	P_gap[((int)('X' - 88)/-43)][((int)('X' - 88)/-43)] = -0.7071;
	P_gap[((int)('X' - 88)/-43)][((int)('-' - 88)/-43)] = 0.7071;
	P_gap[((int)('-' - 88)/-43)][((int)('X' - 88)/-43)] = 0.7071;
	P_gap[((int)('-' - 88)/-43)][((int)('-' - 88)/-43)] = 0.7071;
	
	P1_gap[((int)('X' - 88)/-43)][((int)('X' - 88)/-43)] = -0.7071;
	P1_gap[((int)('X' - 88)/-43)][((int)('-' - 88)/-43)] = 0.7071;
	P1_gap[((int)('-' - 88)/-43)][((int)('X' - 88)/-43)] = 0.7071;
	P1_gap[((int)('-' - 88)/-43)][((int)('-' - 88)/-43)] = 0.7071;
	
	double * sum_letters_gap;
	sum_letters_gap = new double[2];
	
	sum_letters_gap[((int)('X' - 88)/-43)] = 0.98219083;
	sum_letters_gap[((int)('-' - 88)/-43)] = 0.01780917;
	
	//##########################################################################
	// We prescan the sequence to aggregate the gaps.
	//##########################################################################
	
	string prev_seg = "";
	int current_length = 0;
	int scalar_prev_segs = 0;
	
	string prev_segs[gen.getSequenceByID(0).getSequence().length()];
	int * curr_lengths;
	curr_lengths = new int[gen.getSequenceByID(0).getSequence().length()];
	
	int * gap_index;
	gap_index = new int[gen.getSequenceByID(0).getSequence().length()+1]; // Keeps in memory which scalar_prev_seqs belongs to the sequence.
	
	for(int i = 0;i < gen.getSequenceByID(0).getSequence().length();++i){
		string col = "";
		for(int j = 0;j < parent;++j){
			if(nodes.getNodeByID(j).isParent() == 0){
				if(gen.getSequenceByID(j).getSequence().substr(i,1) == "-"){
					col += "-";
				}
				else{
					col += "X";
				}
			}
		}
		if(col == prev_seg){
			++current_length;
		}
		else{
			if(prev_seg != ""){
				prev_segs[scalar_prev_segs] = prev_seg;
				curr_lengths[scalar_prev_segs] = current_length;
				++scalar_prev_segs;
			}
			prev_seg = col;
			current_length = 1;
		}
		gap_index[i] = scalar_prev_segs;
	}

	prev_segs[scalar_prev_segs] = prev_seg;
	curr_lengths[scalar_prev_segs] = current_length;
	gap_index[gen.getSequenceByID(0).getSequence().length()] = scalar_prev_segs;
	//##########################################################################
	// Process
	//#############################################################################
	//double protein_likelihood;
	// To accelerate the process, we will add the observed columns in memory. 
	
	double success = 0;
	double * S;
	double * S2;
	
	double * S_gap;
	double * S_gap2;
	
	int seq_1_length = gen.getSequenceByID(0).getSequence().length();
	int node_list_size = nodes.getNodeListSize();
	if(stop > seq_1_length - (window_size - 1)){
		stop = seq_1_length - (window_size - 1);	
	}
	S = new double[parent * 26];
	S2 = new double[parent * 26];
	S_gap = new double[parent * 2];
	S_gap2 = new double[parent * 2];
	for(int o = start - 1;o< stop;++o){
		int start_window = o;
		int stop_window = o + window_size;
		
		double distance_calc;
		if(success != 0){
			if(window_size<23){
				distance_calc = success *window_size/30;
			}
			else{
				distance_calc = 0.01; // To fix.	
			}
		}
		else{
			distance_calc = 0.01;	
		}
		if(distance_calc < 0.01){
			distance_calc = 0.01;	
		}
	
		
		success = 0;
		int steps = 0;
		string extant = "";	
		
		double lambs_1[aalist_size][parent-children_size];
		double lambs_2[aalist_size][parent-children_size];
		double lambs_3[aalist_size][parent-children_size];
		double lambs_4[aalist_size][parent-children_size];
		
		double lambs_gap_1[aalist_size][parent-children_size];
		double lambs_gap_2[aalist_size][parent-children_size];
		double lambs_gap_3[aalist_size][parent-children_size];
		double lambs_gap_4[aalist_size][parent-children_size];
		
		while(success == 0){
			double trial_likelihood = 0;
			double gradient_likelihood = 0;
			
			double trial_gap_likelihood = 0;
			double gradient_gap_likelihood = 0;
			// We do the gap process first
			
			int child_1;
			int child_2;
			double distance_1;
			double distance_2;
			for(int j = children_size;j < parent;++j){
				child_1 = nodes.getNodeByID(j).getChild_1();
				child_2 = nodes.getNodeByID(j).getChild_2();
				distance_1 = nodes.getNodeByID(j).getDistance_1();
				distance_2 = nodes.getNodeByID(j).getDistance_2();
					
				for(int it = 0;it < aalist_size_gap;++it){
					lambs_gap_1[it][j-children_size] = exp(lambda_gap[(int)(aalist_gap[it][0] - 88)/-43] * distance_1 * distance_calc);
					lambs_gap_2[it][j-children_size] = exp(lambda_gap[(int)(aalist_gap[it][0] - 88)/-43] * distance_2 * distance_calc);
					lambs_gap_3[it][j-children_size] = exp(lambda_gap[(int)(aalist_gap[it][0] - 88)/-43] * distance_1 * (distance_calc + 0.01));
					lambs_gap_4[it][j-children_size] = exp(lambda_gap[(int)(aalist_gap[it][0] - 88)/-43] * distance_2 * (distance_calc + 0.01));
				}
			}

			for(int i = gap_index[start_window];i <= gap_index[stop_window];++i){
				
				
				memset(S_gap,'\0',sizeof(double)*(parent*2));
				memset(S_gap2,'\0',sizeof(double)*(parent*2));
				
				extant = "";
				for(int j = 0;j < children_size;++j){
					string tmp = prev_segs[i];
					S_gap[j+ 2 * ((int)(tmp[j] - 88)/-43)] = 1;
					S_gap2[j+ 2 * ((int)(tmp[j] - 88)/-43)] = 1;
					extant += tmp[j];
				}	
				
				// Loop through parents;
				string letters;
				string letters2;
				string letters3;
				double P_1;
				double P_2;
				double P_1_2;
				double P_2_2;
				
				
				for(int j = children_size;j < parent;++j){
					// Get children ID
					child_1 = nodes.getNodeByID(j).getChild_1();
					child_2 = nodes.getNodeByID(j).getChild_2();
					distance_1 = nodes.getNodeByID(j).getDistance_1();
					distance_2 = nodes.getNodeByID(j).getDistance_2();
					
					// Calculate likelihood
					for(int k = 0;k < aalist_size_gap;++k){
						letters = aalist_gap[k];
						// Calculate transition probabilities
						P_1 = 0;
						P_2 = 0;
						
						P_1_2 = 0;
						P_2_2 = 0;
						for(int n = 0;n < aalist_size_gap;++n){
							letters3 = aalist_gap[n];
							
							double p_temp1 = 0;
							double p_temp2 = 0;
							double p_temp3 = 0;
							double p_temp4 = 0;
														
							double prob = 0;
							double prob2 = 0;
							double prob3 = 0;
							double prob4 = 0;
							
							double lamb = 0;
							double lamb2 = 0;
							double lamb3 = 0;
							double lamb4 = 0;
							for(int it = 0;it < aalist_size_gap;++it){
								
								lamb = lambs_gap_1[it][j-children_size];
								lamb2 = lambs_gap_2[it][j-children_size];
								lamb3 = lambs_gap_3[it][j-children_size];
								lamb4 = lambs_gap_4[it][j-children_size];
								
								prob += (P_gap[(int)(letters[0] - 88)/-43][(int)(aalist_gap[it][0] - 88)/-43] * lamb * P1_gap[(int)(aalist_gap[it][0] - 88)/-43][(int)(letters3[0] - 88)/-43]);
								prob2 +=  (P_gap[(int)(letters[0] - 88)/-43][(int)(aalist_gap[it][0] - 88)/-43] * lamb2 * P1_gap[(int)(aalist_gap[it][0] - 88)/-43][(int)(letters3[0] - 88)/-43]);
								prob3 += (P_gap[(int)(letters[0] - 88)/-43][(int)(aalist_gap[it][0] - 88)/-43] * lamb3 * P1_gap[(int)(aalist_gap[it][0] - 88)/-43][(int)(letters3[0] - 88)/-43]);
								prob4 +=  (P_gap[(int)(letters[0] - 88)/-43][(int)(aalist_gap[it][0] - 88)/-43] * lamb4 * P1_gap[(int)(aalist_gap[it][0] - 88)/-43][(int)(letters3[0] - 88)/-43]);
							}
								
							
							p_temp1 = max(0.0,prob);
							p_temp2 = max(0.0,prob2);
							p_temp3 = max(0.0,prob3);
							p_temp4 = max(0.0,prob4);
							
							P_1 += p_temp1 * S_gap[child_1 + 2 * (int)((letters3[0]-88)/-43)];
							P_2 += p_temp2 * S_gap[child_2 + 2 * (int)((letters3[0]-88)/-43)];
							
							P_1_2 += p_temp3 * S_gap2[child_1 + 2 * (int)((letters3[0]-88)/-43)];
							P_2_2 += p_temp4 * S_gap2[child_2 + 2 * (int)((letters3[0]-88)/-43)];
						}
						
						S_gap[j + 2 * (int)((letters[0] - 88)/-43)] = (P_1 * P_2);
						S_gap2[j + 2 * (int)((letters[0] - 88)/-43)] = (P_1_2 * P_2_2);

					}	
				}
				
				double gap_likelihood = 0;
				double gap_likelihood2 = 0;
				for(int k = 0;k < aalist_size_gap;++k){
					letters = aalist_gap[k];
					
					gap_likelihood += S_gap[(node_list_size-1) + 2*(int)((letters[0] - 88)/-43)] * sum_letters_gap[((int)(letters[0] - 88)/-43)];	
					gap_likelihood2 += S_gap2[(node_list_size-1) + 2*(int)((letters[0] - 88)/-43)] * sum_letters_gap[((int)(letters[0] - 88)/-43)];	
				}
				int gap_count = 0;
				int extent_length = extant.length();
				for(int k = 0;k < extent_length;++k){
					letters = extant.substr(k,1);
					if(letters[0] == '-'){
						++gap_count;
					}
				}
				
				gap_likelihood = log(gap_likelihood);
				gap_likelihood2 = log(gap_likelihood2);
				
				trial_gap_likelihood += gap_likelihood;
				gradient_gap_likelihood += gap_likelihood2;
				
				

				if(gap_index[stop_window] - gap_index[start_window] + 1 > window_size){
					break;	
				}
			}		
			
			// We save the substitutions calculated into an array. 
			
			double save[26][26][parent];
			memset(save,'\0',sizeof(double)*(26*26*parent));
			double save2[26][26][parent];
			memset(save2,'\0',sizeof(double)*(26*26*parent));
			double save3[26][26][parent];
			memset(save3,'\0',sizeof(double)*(26*26*parent));
			double save4[26][26][parent];
			memset(save4,'\0',sizeof(double)*(26*26*parent));
			
			// Substitution process
			for(int j = children_size;j < parent;++j){
				
				// Get children ID
				child_1 = nodes.getNodeByID(j).getChild_1();
				child_2 = nodes.getNodeByID(j).getChild_2();
					
				distance_1 = nodes.getNodeByID(j).getDistance_1();
				distance_2 = nodes.getNodeByID(j).getDistance_2();
					
				for(int it = 0;it < aalist_size;++it){
					lambs_1[it][j-children_size] = exp(lambda[aalist[it][0] - 'A'] * distance_1 * distance_calc);
					lambs_2[it][j-children_size] = exp(lambda[aalist[it][0] - 'A'] * distance_2 * distance_calc);
					lambs_3[it][j-children_size] = exp(lambda[aalist[it][0] - 'A'] * distance_1 * (distance_calc + 0.01));
					lambs_4[it][j-children_size] = exp(lambda[aalist[it][0] - 'A'] * distance_2 * (distance_calc + 0.01));
				}

			}
			
			for(int i = start_window;i < stop_window;++i){
				
				memset(S,'\0',sizeof(double)*(parent*26));
				memset(S2,'\0',sizeof(double)*(parent*26));
				int verify = 0;
				for(int j = 0;j < children_size;++j){
					string tmp = nodes.getNodeByID(j).getSequence();
					if(tmp[i] == 'X' || tmp[i] == '-'){
						
					}
					else{
						S[j*26 + (tmp[i] - 'A')] = 1;
						S2[j*26 +  (tmp[i] - 'A')] = 1;
						verify = 1;
					}
					
				}
				if(verify == 0){ continue;}

				// Loop through parents;
				string letters;
				string letters2;
				string letters3;
				double P_1;
				double P_2;
				double P_1_2;
				double P_2_2;
			
				for(int j = children_size;j < parent;++j){
					// Get children ID
					child_1 = nodes.getNodeByID(j).getChild_1();
					child_2 = nodes.getNodeByID(j).getChild_2();
					
					distance_1 = nodes.getNodeByID(j).getDistance_1();
					distance_2 = nodes.getNodeByID(j).getDistance_2();

					// Calculate likelihood
					for(int k = 0;k < aalist_size;++k){
						letters = aalist[k];
						// Calculate transition probabilities
						P_2 = 0;
						P_1 = 0;
						P_1_2 = 0;
						P_2_2 = 0;
						
						for(int n = 0;n < aalist_size;++n){
							letters3 = aalist[n];
							
							double p_temp1 = 0;
							double p_temp2 = 0;
							double p_temp3 = 0;
							double p_temp4 = 0;
							
							if(save[letters[0] - 'A'][letters3[0] - 'A'][j] == 0){
								
								double prob = 0;
								double prob2 = 0;
								double prob3 = 0;
								double prob4 = 0;
								
								double lamb = 0;
								double lamb2 = 0;
								double lamb3 = 0;
								double lamb4 = 0;
								
								
								for(int it = 0;it < aalist_size;++it){
								
									lamb = lambs_1[it][j-children_size];
									lamb2 = lambs_2[it][j-children_size];
									lamb3 = lambs_3[it][j-children_size];
									lamb4 = lambs_4[it][j-children_size];
									
									prob += (P[letters[0] - 'A'][aalist[it][0] - 'A'] * lamb * P1[aalist[it][0] - 'A'][letters3[0] - 'A']);
									prob2 +=  (P[letters[0] - 'A'][aalist[it][0] - 'A'] * lamb2 * P1[aalist[it][0] - 'A'][letters3[0] - 'A']);
									prob3 += (P[letters[0] - 'A'][aalist[it][0] - 'A'] * lamb3 * P1[aalist[it][0] - 'A'][letters3[0] - 'A']);
									prob4 +=  (P[letters[0] - 'A'][aalist[it][0] - 'A'] * lamb4 * P1[aalist[it][0] - 'A'][letters3[0] - 'A']);
								}
								
								p_temp1 = max(0.0,prob);
								p_temp2 = max(0.0,prob2);
								p_temp3 = max(0.0,prob3);
								p_temp4 = max(0.0,prob4);

								// Save computed values
								
								save[letters[0] - 'A'][letters3[0] - 'A'][j] = p_temp1;
								save2[letters[0] - 'A'][letters3[0] - 'A'][j] = p_temp2;
								save3[letters[0] - 'A'][letters3[0] - 'A'][j] = p_temp3;
								save4[letters[0] - 'A'][letters3[0] - 'A'][j] = p_temp4;

								// Model is reversible, therefore we can save the reverse values.

								save[letters3[0] - 'A'][letters[0] - 'A'][j] = p_temp1;
								save2[letters3[0] - 'A'][letters[0] - 'A'][j] = p_temp2;
								save3[letters3[0] - 'A'][letters[0] - 'A'][j] = p_temp3;
								save4[letters3[0] - 'A'][letters[0] - 'A'][j] = p_temp4;

								
								
							}
							else{
								
								p_temp1 = save[letters[0] - 'A'][letters3[0] - 'A'][j];
								p_temp2 = save2[letters[0] - 'A'][letters3[0] - 'A'][j];
								p_temp3 = save3[letters[0] - 'A'][letters3[0] - 'A'][j];
								p_temp4 = save4[letters[0] - 'A'][letters3[0] - 'A'][j];
								
							}
							
							
							P_1 += p_temp1 * S[child_1 *26 + (letters3[0]-'A')];
							P_2 += p_temp2 * S[child_2 *26 + (letters3[0]-'A')];
							
							P_1_2 += p_temp3 * S2[child_1 *26  +  (letters3[0]-'A')];
							P_2_2 += p_temp4 * S2[child_2 *26 +(letters3[0]-'A')];
						}
						
						if(P_2 == 0){
							P_2 = 1;	
						}
						if(P_1 == 0){
							P_1 = 1;	
						}
						if(P_1_2 == 0){
							P_1_2 = 1;	
						}
						if(P_2_2 == 0){
							P_2_2 = 1;	
						}
						
						if(P_1 == 1 && P_2 == 1){
							P_1 = 0;
							P_2 = 0;
						}
						if(P_1_2 == 1 && P_2_2 == 1){
							P_1_2 = 0;
							P_2_2 = 0;
						}

						S[j*26 + (letters[0] - 'A')] = (P_1 * P_2);
						S2[j*26 + (letters[0] - 'A')] = (P_1_2 * P_2_2);
						
					}

				}
				// Compute final likelihood
				// Compute maximal likelihood
				double likelihood = 0;
				double L = 0;
				double likelihood2 = 0;
				double L2 = 0;
				for(int k = 0;k < aalist_size;++k){
					letters = aalist[k];
					
					L = S[(node_list_size-1)*26 + (letters[0] - 'A')] * sum_letters[letters[0]-'A'];
					L2 = S2[(node_list_size-1) *26 +  (letters[0] - 'A')] * sum_letters[letters[0]-'A'];
					likelihood += L;
					likelihood2 += L2;
					
				}
				
				trial_likelihood += log(likelihood);
				gradient_likelihood += log(likelihood2);				
			}

			trial_likelihood += trial_gap_likelihood;
			gradient_likelihood += gradient_gap_likelihood;
			
			if(gradient_likelihood<trial_likelihood || (gradient_likelihood - trial_likelihood<0.01)){
				if(gradient_likelihood > trial_likelihood){
					success = distance_calc + 0.01;
				}
				else{
					success = distance_calc;
				}
			}
			else{
				if(gradient_likelihood - trial_likelihood > 1){
					distance_calc += sqrt(gradient_likelihood - trial_likelihood)/10;
				}
				else if(gradient_likelihood - trial_likelihood > 0.1){
					distance_calc += sqrt(gradient_likelihood - trial_likelihood)/3;
				}
				else{
					distance_calc += sqrt(gradient_likelihood - trial_likelihood)/5;
				}
			}
		}
		if(o == 0){
			for(int y = 0;y<((window_size-1)/2);++y){
				cout << y+1 << "	" << success << endl;	
			}
		}
		cout << o +((window_size+1)/2) << "	" << success << endl;
		if(o == seq_1_length-window_size){
			for(int y = seq_1_length-((window_size-3)/2);y<=seq_1_length;++y){
				cout << y << "	" << success << endl;	
			}
		}
		
	}
	delete [] S_gap;
	delete [] S_gap2;
	delete [] S;
	delete [] S2;
	delete[] sum_letters;
	delete[] lambda;
	return 0;
	
}
